// SPDX-License-Identifier: Unlicense

#include "camera.h"

#include "global.h"
#include "opengl.h"
#include "shader.h"
#include "helper.h"

static void calcCameraTarget(const ManifestIndex selectedManifestIndex, vec3& cameraTarget)
{
  { // calc the camera Target
    const i32 currentAnchorIndex = Song::getCurrentAnchorIndex(selectedManifestIndex);

    const Song::Anchor& anchor0 = Global::songTrackLevelAdjusted[selectedManifestIndex].anchors[currentAnchorIndex];

    i8 left = anchor0.fret;
    i8 right = anchor0.fret + anchor0.width;

    for (i32 i = currentAnchorIndex + 1; i < i32(Global::songTrackLevelAdjusted[selectedManifestIndex].anchors.size()); ++i)
    {
      const Song::Anchor& anchor1 = Global::songTrackLevelAdjusted[selectedManifestIndex].anchors[i];

      const TimeNS noteTime1 = -anchor1.timeNS + Global::musicTimeElapsedNS;

      if (noteTime1 < timeNS_From_Seconds(-Settings::cameraLinearAnchorTackingDuration))
        break;

      left = min_(left, anchor1.fret);
      right = max_(right, i8(anchor1.fret + anchor1.width));
    }

    i32 width = right - left;
    cameraTarget.x = -Settings::cameraLinearXFactor * f32(left - 1 + right - 1) * 0.5f - Settings::cameraLinearXOffset;
    cameraTarget.y = -Settings::cameraLinearYFactor * f32(width) * 0.5f - Settings::cameraLinearYOffset;
    cameraTarget.z = -Settings::cameraLinearZFactor * f32(width) * 0.5f - Settings::cameraLinearZOffset;
  }
}

static void tickCameraMovementFixed(vec3& cameraPosition)
{
  cameraPosition.x = Settings::cameraFixedXOffset;
  cameraPosition.y = Settings::cameraFixedYOffset;
  cameraPosition.z = Settings::cameraFixedZOffset;
}

static void tickCameraMovementLinear(const vec3& desiredDirection, vec3& cameraPosition)
{
  const vec3 framePosChange = vec::multiply(desiredDirection, TimeNS_To_Seconds(Global::frameDelta));
  cameraPosition.x += framePosChange.x;
  cameraPosition.y += framePosChange.y;
  cameraPosition.z += framePosChange.z;
}

// TODO: this does not work. Idea is to to have 
static void tickCameraMovementKinematic(const vec3& desiredDirection, vec3& cameraPosition, const vec3& cameraTarget)
{
  static bool arrived = false;
  static bool arrivedLastFrame = false;
  vec3 desiredDirection2;
  {
    const f32 cameraDesiredDirectionLength = vec::length(desiredDirection);
    if (cameraDesiredDirectionLength < Settings::cameraLinearBreakRadius)
    {
      desiredDirection2 = vec::multiply(vec::norm(desiredDirection), Settings::cameraLinearMaximumVelocity * (cameraDesiredDirectionLength / Settings::cameraLinearBreakRadius));
      arrivedLastFrame = arrived;
      arrived = true;
    }
    else
    {
      desiredDirection2 = vec::multiply(vec::norm(desiredDirection), Settings::cameraLinearMaximumVelocity);
      arrived = false;
    }
  }

  static vec3 velocity;
  vec3 steering;
  steering.x = desiredDirection2.x - velocity.x;
  steering.y = desiredDirection2.y - velocity.y;
  steering.z = desiredDirection2.z - velocity.z;

  static f32 cameraDesiredDirectionLengthLastFrame = F32::max;
  static bool skip;
  if (arrived)
  {
    if (arrivedLastFrame)
    { // add check to detect if breaking was so hard that we are going backwards now
      if (skip || vec::length(desiredDirection2) > cameraDesiredDirectionLengthLastFrame)
      {
        cameraPosition.x = cameraTarget.x;
        cameraPosition.y = cameraTarget.y;
        cameraPosition.z = cameraTarget.z;
        skip = true;
        return;
      }
    }
    else
    {
      steering = vec::multiply(vec::norm(steering), Settings::cameraLinearMaximumBreakForce);
    }
  }
  else
  {
    steering = vec::truncate(steering, Settings::cameraLinearMaximumForce);
  }
  skip = false;
  cameraDesiredDirectionLengthLastFrame = vec::length(desiredDirection2);

  velocity.x += steering.x * TimeNS_To_Seconds(Global::frameDelta);
  velocity.y += steering.y * TimeNS_To_Seconds(Global::frameDelta);
  velocity.z += steering.z * TimeNS_To_Seconds(Global::frameDelta);
  velocity = vec::truncate(velocity, Settings::cameraLinearMaximumVelocity);

  const vec3 framePosChange = vec::multiply(velocity, TimeNS_To_Seconds(Global::frameDelta));
  cameraPosition.x += framePosChange.x;
  cameraPosition.y += framePosChange.y;
  cameraPosition.z += framePosChange.z;
}

#ifdef SHR3D_OPENXR_PCVR
static mat4 updateCameraXrFixedPosition(const vec3& cameraPosition)
{
  const mat4 rotMatX{
    .m11 = cosf(Settings::cameraXrFixedXRotation),
    .m21 = sinf(Settings::cameraXrFixedXRotation),
    .m12 = -sinf(Settings::cameraXrFixedXRotation),
    .m22 = cosf(Settings::cameraXrFixedXRotation)
  };

  const mat4 rotMatY{
    .m00 = cosf(Settings::cameraXrFixedYRotation),
    .m20 = sinf(Settings::cameraXrFixedYRotation),
    .m02 = -sinf(Settings::cameraXrFixedYRotation),
    .m22 = cosf(Settings::cameraXrFixedYRotation)
  };

  const mat4 translation{
    .m03 = cameraPosition.x,
    .m13 = cameraPosition.y,
    .m23 = cameraPosition.z
  };

  return vec::multiply(vec::multiply(rotMatX, rotMatY), translation);
}
#endif // SHR3D_OPENXR_PCVR

static mat4 updateCameraFixedPosition(const vec3& cameraPosition)
{
  const mat4 rotMatX{
    .m11 = cosf(Settings::cameraFixedXRotation),
    .m21 = sinf(Settings::cameraFixedXRotation),
    .m12 = -sinf(Settings::cameraFixedXRotation),
    .m22 = cosf(Settings::cameraFixedXRotation)
  };

  const mat4 rotMatY{
    .m00 = cosf(Settings::cameraFixedYRotation),
    .m20 = sinf(Settings::cameraFixedYRotation),
    .m02 = -sinf(Settings::cameraFixedYRotation),
    .m22 = cosf(Settings::cameraFixedYRotation)
  };

  const mat4 translation{
    .m03 = cameraPosition.x,
    .m13 = cameraPosition.y,
    .m23 = cameraPosition.z
  };

  return vec::multiply(vec::multiply(rotMatX, rotMatY), translation);
}

static mat4 updateCameraLinearPosition(const vec3& cameraPosition)
{
  mat4 rotMatX;
  rotMatX.m11 = cosf(Settings::cameraLinearXRotation);
  rotMatX.m21 = sinf(Settings::cameraLinearXRotation);
  rotMatX.m12 = -sinf(Settings::cameraLinearXRotation);
  rotMatX.m22 = cosf(Settings::cameraLinearXRotation);

  mat4 rotMatY;
  rotMatY.m00 = cosf(Settings::cameraLinearYRotation);
  rotMatY.m20 = sinf(Settings::cameraLinearYRotation);
  rotMatY.m02 = -sinf(Settings::cameraLinearYRotation);
  rotMatY.m22 = cosf(Settings::cameraLinearYRotation);

  mat4 translation;
  translation.m03 = cameraPosition.x;
  translation.m13 = cameraPosition.y;
  translation.m23 = cameraPosition.z;

  return vec::multiply(vec::multiply(rotMatX, rotMatY), translation);
}

static mat4 noclipMovement(const vec3& cameraTarget)
{
  static vec3 noclipPosition = cameraTarget;

  if (Global::inputD.heldDown)
    noclipPosition.x -= Settings::highwayScrollSpeed * TimeNS_To_Seconds(Global::frameDelta);
  else if (Global::inputA.heldDown)
    noclipPosition.x += Settings::highwayScrollSpeed * TimeNS_To_Seconds(Global::frameDelta);
  if (Global::inputW.heldDown)
    noclipPosition.z += Settings::highwayScrollSpeed * TimeNS_To_Seconds(Global::frameDelta);
  else if (Global::inputS.heldDown)
    noclipPosition.z -= Settings::highwayScrollSpeed * TimeNS_To_Seconds(Global::frameDelta);
  if (Global::inputE.heldDown)
    noclipPosition.y -= Settings::highwayScrollSpeed * TimeNS_To_Seconds(Global::frameDelta);
  else if (Global::inputC.heldDown)
    noclipPosition.y += Settings::highwayScrollSpeed * TimeNS_To_Seconds(Global::frameDelta);

  mat4 translation;
  translation.m03 = noclipPosition.x;
  translation.m13 = noclipPosition.y;
  translation.m23 = noclipPosition.z;

  static i32 initialPosX = Global::inputCursorPosX;
  const f32 rotX = (initialPosX - Global::inputCursorPosX) * 0.004f;

  static i32 initialPosY = Global::inputCursorPosY;
  const f32 rotY = (initialPosY - Global::inputCursorPosY) * -0.004f;

  mat4 rotMatX;
  rotMatX.m11 = cosf(rotY);
  rotMatX.m21 = sinf(rotY);
  rotMatX.m12 = -sinf(rotY);
  rotMatX.m22 = cosf(rotY);

  mat4 rotMatY;
  rotMatY.m00 = cosf(rotX);
  rotMatY.m20 = sinf(rotX);
  rotMatY.m02 = -sinf(rotX);
  rotMatY.m22 = cosf(rotX);

  return vec::multiply(vec::multiply(rotMatX, rotMatY), translation);
}

mat4 Camera::tickCalculateViewMat(const ManifestIndex selectedManifestIndex, vec3& cameraPosition, vec3& cameraTarget)
{
  static bool skipTick = true; // when a song is loaded the game freezes for a few sec. Don't tick with such a high Global::frameDelta. Camera positon will be off.
  if (Global::selectedSongIndex == -1 || Global::musicTimeElapsedNS > (Global::songInfos[Global::selectedSongIndex].songLength + Global::songInfos[Global::selectedSongIndex].shredDelayEnd))
  {
    skipTick = true;
    return {};
  }
  else if (skipTick)
  {
    skipTick = false;
    return {};
  }

  calcCameraTarget(selectedManifestIndex, cameraTarget);

  const vec3 desiredDirection =
  {
    .x = cameraTarget.x - cameraPosition.x,
    .y = cameraTarget.y - cameraPosition.y,
    .z = cameraTarget.z - cameraPosition.z,
  };

  switch (Settings::cameraMode)
  {
  case CameraMode::fixed:
    tickCameraMovementFixed(cameraPosition);
    break;
  case CameraMode::linear:
    tickCameraMovementLinear(desiredDirection, cameraPosition);
    break;
  case CameraMode::kinematic:
    tickCameraMovementKinematic(desiredDirection, cameraPosition, cameraTarget);
    break;
  }

  if (Global::inputNoclip.toggled)
    return noclipMovement(cameraTarget);

#ifdef SHR3D_OPENXR_PCVR
  if (Settings::xrEnabled)
  {
    return updateCameraXrFixedPosition(cameraPosition);
  }
  else
#endif // SHR3D_OPENXR_PCVR
  {
    switch (Settings::cameraMode)
    {
    case CameraMode::fixed:
      return updateCameraFixedPosition(cameraPosition);
    case CameraMode::linear:
      return updateCameraLinearPosition(cameraPosition);
    case CameraMode::kinematic:
      return updateCameraLinearPosition(cameraPosition);
    }
  }

  unreachable();
}

//mat4 Camera::tickCalculateViewMat2(const ManifestIndex selectedManifestIndex, vec3& cameraPosition, vec3& cameraTarget)
//{
//  //static bool skipTick = true; // when a song is loaded the game freezes for a few sec. Don't tick with such a high Global::frameDelta. Camera positon will be off.
//  //if (Global::selectedSongIndex == -1 || Global::musicTimeElapsedNS > (Global::songInfos[Global::selectedSongIndex].songLength + Global::songInfos[Global::selectedSongIndex].shredDelay))
//  //{
//  //  skipTick = true;
//  //  return {};
//  //}
//  //else if (skipTick)
//  //{
//  //  skipTick = false;
//  //  return {};
//  //}
//
//  cameraTarget.x = -11.7;
//  cameraTarget.y = -7.8;
//  cameraTarget.z = -8.7;
//
//  //calcCameraTarget(selectedManifestIndex, cameraTarget);
//
//  const vec3 desiredDirection =
//  {
//    .x = cameraTarget.x - cameraPosition.x,
//    .y = cameraTarget.y - cameraPosition.y,
//    .z = cameraTarget.z - cameraPosition.z,
//  };
//
//  switch (Settings::cameraMode)
//  {
//  case CameraMode::fixed:
//    tickCameraMovementFixed(cameraPosition);
//    break;
//  case CameraMode::linear:
//    tickCameraMovementLinear(desiredDirection, cameraPosition);
//    break;
//  case CameraMode::kinematic:
//    tickCameraMovementKinematic(desiredDirection, cameraPosition, cameraTarget);
//    break;
//  }
//
//  if (Global::inputNoclip.toggled)
//    return noclipMovement(cameraTarget);
//
//  switch (Settings::cameraMode)
//  {
//  case CameraMode::fixed:
//    return updateCameraFixedPosition(cameraPosition);
//  case CameraMode::linear:
//    return updateCameraLinearPosition(cameraPosition);
//  case CameraMode::kinematic:
//    return updateCameraLinearPosition(cameraPosition);
//  }
//
//  ASSERT(false);
//  return {};
//}

mat4 Camera::tickCalculateProjectionViewMat(const i32 resolutionWidth, const i32 resolutionHeight, const mat4& viewMat)
{
  mat4 projectionMat;
  const f32 aspect = Settings::graphicsScaling == ScalingMode::HorPlus
    ? f32(resolutionWidth) / f32(resolutionHeight)
    : 16.0f / 9.0f;
  {
    const f32 near_ = 0.1f;
    const f32 far_ = Settings::highwayViewDistance * 2.0f; // * 2.0f is there to not clip off drawing when moving back a lot in VR.
    const f32 fieldOfView = Settings::cameraMode == CameraMode::fixed ? Settings::cameraFixedFieldOfView : Settings::cameraLinearFieldOfView;
    const f32 divisor = tanf((fieldOfView * PI_) / 360.0f);
    projectionMat.m00 = 1.0f / (aspect * divisor);
    projectionMat.m11 = 1.0f / divisor;
    projectionMat.m22 = -(far_ + near_) / (far_ - near_);
    projectionMat.m23 = -(2.0f * far_ * near_) / (far_ - near_);
    projectionMat.m32 = -1.0f;
    projectionMat.m33 = 0.0f;
  }

  mat4 viewProjectionMat = vec::multiply(projectionMat, viewMat);

  return viewProjectionMat;
}

//mat4 Camera::tickCalculateProjectionViewMat2(const i32 resolutionWidth, const i32 resolutionHeight, const mat4& viewMat)
//{
//  mat4 projectionMat;
//  const f32 aspect = Settings::graphicsScaling == ScalingMode::HorPlus
//    ? f32(resolutionWidth) / f32(resolutionHeight)
//    : 16.0f / 9.0f;
//  {
//    const f32 near_ = 0.1f;
//    const f32 far_ =
//#ifdef SHR3D_ENVIRONMENT_MODEL
//      !Settings::environmentModel.empty() ? max_(Global::environmentModelBoundingZ * Settings::environmentModelScale, Settings::highwayViewDistance) :
//#endif // SHR3D_ENVIRONMENT_MODEL
//      Settings::highwayViewDistance;
//    const f32 fieldOfView = Settings::cameraMode == CameraMode::fixed ? Settings::cameraFixedFieldOfView : Settings::cameraLinearFieldOfView;
//    const f32 divisor = tanf((fieldOfView * PI_) / 360.0f);
//    projectionMat.m00 = Global::debugValue[1] + Global::debugValue[0] * (1.0f / (aspect * divisor));
//    projectionMat.m11 = Global::debugValue[1] + Global::debugValue[0] * (1.0f / divisor);
//    projectionMat.m22 = Global::debugValue[0] * (-(far_ + near_) / (far_ - near_));
//    projectionMat.m23 = Global::debugValue[0] * (-(2.0f * far_ * near_) / (far_ - near_));
//    projectionMat.m32 = Global::debugValue[0] * (-1.0f);
//    projectionMat.m33 = Global::debugValue[1] + 0.0f;
//  }
//
//  mat4 viewProjectionMat = vec::multiply(projectionMat, viewMat);
//
//  return viewProjectionMat;
//}

static void drawCircle()
{
  const i32 circleSegments = 32;
  const f32 r = 0.1f;
  GLfloat v[circleSegments * 5];
  for (i32 i = 0; i < circleSegments; ++i)
  {
    const f32 alpha = 2.0f * PI_ * f32(i) / f32(circleSegments);
    v[i * 5] = r * cosf(alpha); // x
    v[i * 5 + 1] = r * sinf(alpha); // y
    v[i * 5 + 2] = 0.0f; // z
    v[i * 5 + 3] = 0.0f; // u
    v[i * 5 + 4] = 0.0f; // v
  }

  GL(glBufferData(GL_ARRAY_BUFFER, sizeof(v), v, GL_STATIC_DRAW));
  GL(glDrawArrays(GL_TRIANGLE_FAN, 0, circleSegments));
}

void Camera::render(const mat4& viewProjectionMat, const vec3& cameraPosition, const vec3& cameraTarget)
{
  if (!Global::inputNoclip.toggled)
    return;

  GL(glUseProgram(Shader::world));

  {
    mat4 model;
    model.m03 = -cameraTarget.x;
    model.m13 = -cameraTarget.y;
    model.m23 = -cameraTarget.z;

    const mat4 mvp = vec::multiply(viewProjectionMat, model);
    GL(glUniformMatrix4fv(Shader::worldUniformModelViewProjection, 1, GL_FALSE, &mvp.m00));

    GL(glUniform4f(Shader::worldUniformColor, 1.0f, 0.0f, 0.0f, 1.0f));
    drawCircle();
  }

  {
    mat4 model;
    model.m03 = -cameraPosition.x;
    model.m13 = -cameraPosition.y;
    model.m23 = -cameraPosition.z;

    const mat4 mvp = vec::multiply(viewProjectionMat, model);
    GL(glUniformMatrix4fv(Shader::worldUniformModelViewProjection, 1, GL_FALSE, &mvp.m00));

    GL(glUniform4f(Shader::worldUniformColor, 0.0f, 1.0f, 0.0f, 1.0f));
    drawCircle();
  }
}
