// SPDX-License-Identifier: Unlicense

#include "model.h"

#ifdef SHR3D_ENVIRONMENT_MODEL

#include "file.h"
#include "global.h"
#include "opengl.h"
#include "shader.h"
#include "string_.h"
#include "texture.h"

void Model::init()
{
  const std::vector<std::u8string> modelFilepaths = File::filesInDirectory(Settings::pathModel.c_str());

  for (const std::u8string& filepath : modelFilepaths)
  {
    if (File::extension(filepath.c_str()) != std::u8string(u8".3d"))
      continue;

    {
      const std::u8string modelName = File::filename(filepath.c_str());
      Global::environmentModelNames.push_back(modelName);
    }
  }
}

void Model::tick()
{
  static std::u8string activeModels[Const::environmentModelMaxCount];

  for (i8 i = 0; i < Const::environmentModelMaxCount; ++i)
  {
    if (activeModels[i] == Settings::environmentModel[i])
      continue;

    activeModels[i] = Settings::environmentModel[i];

    if (Global::environmentModelIndexData[i].size() != 0)
    {
      Global::environmentModelIndexData[i].clear();
      Global::environmentModelVertexData[i].clear();
      GL(glDeleteTextures(1, &Global::modelTexture[i]));
    }

    if (!Settings::environmentModel[i].empty())
    {
      const std::u8string filepath = Settings::pathModel + Settings::environmentModel[i];
      const std::u8string textureFilepath = File::replace_extension(filepath.c_str(),
#ifdef __ANDROID__
        u8".astc"
#else // __ANDROID__
        u8".dds"
#endif // __ANDROID__
      );

      if (File::exists(filepath.c_str()) && File::exists(textureFilepath.c_str()))
      {
        {
          FILE* file = File::fopen_u8(filepath.c_str(), true);
          ASSERT(file != nullptr);

          struct
          {
            u32 magic;
            u32 version;
            u32 vertexCount;
            u32 triangleCount;
          } header;
          fread(&header, sizeof(header), 1, file);

          ASSERT(header.magic == 3199882682);

          switch (header.version)
          {
          case 0:
            ASSERT(header.vertexCount != 0);
            ASSERT(header.triangleCount == 0);
            Global::environmentModelVertexData[i].resize(5 * header.vertexCount);
            fread(Global::environmentModelVertexData[i].data(), sizeof(f32), Global::environmentModelVertexData[i].size(), file);
            ASSERT(ftell(file) == sizeof(header) + Global::environmentModelVertexData[i].size() * sizeof(f32));
            break;
          case 1:
            ASSERT(header.vertexCount != 0);
            ASSERT(header.triangleCount != 0);
            Global::environmentModelVertexData[i].resize(5 * header.vertexCount);
            fread(Global::environmentModelVertexData[i].data(), sizeof(f32), Global::environmentModelVertexData[i].size(), file);
            ASSERT(ftell(file) == sizeof(header) + Global::environmentModelVertexData[i].size() * sizeof(f32));
            Global::environmentModelIndexData[i].resize(3 * header.triangleCount);
            fread(Global::environmentModelIndexData[i].data(), sizeof(u32), Global::environmentModelIndexData[i].size(), file);
            ASSERT(ftell(file) == sizeof(header) + Global::environmentModelVertexData[i].size() * sizeof(f32) + Global::environmentModelIndexData[i].size() * sizeof(u32));
            break;
          default:
            unreachable();
          }

          fclose(file);
        }
        {
          const std::vector<u8> textureData = File::read(textureFilepath.c_str());
          Global::modelTexture[i] = Texture::openGlLoadTexture(textureData.data(), i32(textureData.size()));
          ASSERT(Global::modelTexture[i] != 0);
        }
      }
    }
  }
}

void Model::render(const mat4& viewProjectionMat)
{
  i8 firstUsedModel = 0;
  for (; firstUsedModel < Const::environmentModelMaxCount; ++firstUsedModel)
  {
    if (!Settings::environmentModel[firstUsedModel].empty())
      break;
  }

  if (firstUsedModel == Const::environmentModelMaxCount)
    return;

  GL(glUseProgram(Shader::model));

  for (i8 i = 0; i < Const::environmentModelMaxCount; ++i)
  {
    if (Global::environmentModelVertexData[i].size() == 0)
      continue;

    mat4 model;
    {
      const mat4 scale{
        .m00 = Settings::environmentModelScale[i],
        .m11 = Settings::environmentModelScale[i],
        .m22 = Settings::environmentModelScale[i]
      };

      const f32 rotation = Settings::environmentModelRotation[i] * 2.0f * PI_;
      const mat4 rotate{
        .m00 = cosf(rotation),
        .m20 = sinf(rotation),
        .m02 = -sinf(rotation),
        .m22 = cosf(rotation)
      };

      model = vec::multiply(rotate, scale);
    }
    {
      const mat4 transform
      {
        .m03 = Settings::environmentModelX[i],
        .m13 = Settings::environmentModelY[i],
        .m23 = Settings::environmentModelZ[i]
      };

      model = vec::multiply(transform, model);
    }

    const mat4 mvp = vec::multiply(viewProjectionMat, model);

    GL(glUniformMatrix4fv(Shader::modelUniformModelViewProjection, 1, GL_FALSE, &mvp.m00));

    GL(glBindTexture(GL_TEXTURE_2D, Global::modelTexture[i]));

    GL(glBufferData(GL_ARRAY_BUFFER, Global::environmentModelVertexData[i].size() * sizeof(f32), Global::environmentModelVertexData[i].data(), GL_STATIC_DRAW));
    if (Global::environmentModelIndexData[i].size() == 0)
    {
      GL(glDrawArrays(GL_TRIANGLES, 0, Global::environmentModelVertexData[i].size() / 5));
    }
    else
    {
      GL(glBufferData(GL_ELEMENT_ARRAY_BUFFER, Global::environmentModelIndexData[i].size() * sizeof(u32), Global::environmentModelIndexData[i].data(), GL_STATIC_DRAW));
      GL(glDrawElements(GL_TRIANGLES, Global::environmentModelIndexData[i].size(), GL_UNSIGNED_INT, 0));
    }
  }
  GL(glBindTexture(GL_TEXTURE_2D, Global::texture));
}

#endif // SHR3D_ENVIRONMENT_MODEL
