// SPDX-License-Identifier: Unlicense

#include "sfxPluginClap.h"

#ifdef SHR3D_SFX_PLUGIN_CLAP

#include "../deps/clap/include/clap.h"
#include "global.h"
#include "settings.h"
#include "string_.h"
#include "version.h"
#include <filesystem>

#ifdef SHR3D_WINDOW_SDL
#include <SDL.h>
#include <SDL_syswm.h>
#endif // SHR3D_WINDOW_SDL

#if defined(PLATFORM_WINDOWS)
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>
typedef HMODULE dlhandle;
typedef HWND ClapWindow;
#define dlopen(x, y) LoadLibrary(x)
#define dlsym(x, y) (void*)GetProcAddress(x, y)
#else // PLATFORM_WINDOWS
#include <dlfcn.h>
typedef ::Window ClapWindow; // there is a "class Window" in rubberband
typedef void* dlhandle;
#endif // PLATFORM_WINDOWS


#ifdef PLATFORM_WINDOWS
#define CLAP_WINDOW_API_OS CLAP_WINDOW_API_WIN32
#else // PLATFORM_WINDOWS
#define CLAP_WINDOW_API_OS CLAP_WINDOW_API_X11
#endif // PLATFORM_WINDOWS


static std::mutex loadSfxPluginClapInstanceMutex;

static const void* clapGetExtension(const clap_host_t* host, const char* extension_id)
{
  return nullptr;
}

static void clapRequestRestart(const clap_host_t* host)
{
}

static void clapRequestProcess(const clap_host_t* host)
{
}

static void clapRequestCallback(const clap_host_t* host)
{
}

static const clap_host host{
  .clap_version = CLAP_VERSION,
  .host_data = nullptr,
  .name = "Shr3D",
  .vendor = "Shr3D Industries",
  .url = "https://shr3d.app/",
  .version = VERSION_STR,
  .get_extension = clapGetExtension,
  .request_restart = clapRequestRestart,
  .request_process = clapRequestProcess,
  .request_callback = clapRequestCallback
};

struct EventList
{
  static constexpr u32 heapSize = 1024 * 1024;
  u8* heap;
  u8* tail;
  std::vector<u32> eventSize;

  EventList()
    : heap((u8*)malloc(heapSize))
    , tail(heap)
  {
  }
};

static u32 clapSize(const struct clap_input_events* list)
{
  EventList* eventList = static_cast<EventList*>(list->ctx);
  return eventList->eventSize.size();
}

static const clap_event_header_t* clapGet(const struct clap_input_events* list, uint32_t index)
{
  EventList* eventList = static_cast<EventList*>(list->ctx);

  eventList->tail -= eventList->eventSize[index];
  eventList->eventSize.pop_back();
  const clap_event_header* header = reinterpret_cast<const clap_event_header*>(eventList->tail);

  return header;
}

static bool clapTryPush(const struct clap_output_events* list, const clap_event_header* event)
{
  EventList* eventList = static_cast<EventList*>(list->ctx);

  ASSERT(eventList->tail - eventList->heap + event->size <= EventList::heapSize);

  memcpy(eventList->tail, event, event->size);
  eventList->eventSize.push_back(event->size);
  eventList->tail += event->size;

  return true;
}

struct PluginInstance
{
  const clap_plugin_t* plugin;
  bool sheduleEventFlush = false;

  clap_audio_buffer* audio_inputs = new clap_audio_buffer
  {
    .data32 = nullptr,
    .data64 = nullptr,
    .channel_count = 2,
    .latency = 0,
    .constant_mask = 0
  };

  clap_audio_buffer* audio_outputs = new clap_audio_buffer
  {
    .data32 = nullptr,
    .data64 = nullptr,
    .channel_count = 2,
    .latency = 0,
    .constant_mask = 0
  };

  const clap_input_events* in_events = new clap_input_events
  {
    .ctx = new EventList,
    .size = clapSize,
    .get = clapGet
  };
  const clap_output_events* out_events = new clap_output_events
  {
    .ctx = in_events->ctx,
    .try_push = clapTryPush
  };

  clap_process* process = new clap_process
  {
    .steady_time = 0,
    .frames_count = 0,
    .transport = nullptr,
    .audio_inputs = audio_inputs,
    .audio_outputs = audio_outputs,
    .audio_inputs_count = 1,
    .audio_outputs_count = 1,
    .in_events = in_events,
    .out_events = out_events
  };
};

struct ClapPlugin
{
  std::vector<PluginInstance> pluginInstances;
  const clap_plugin_factory* pluginFactory;

#ifdef SHR3D_WINDOW_SDL
  SDL_Window* window;
#endif // SHR3D_WINDOW_SDL
#ifdef SHR3D_WINDOW_WIN32
  HWND window;
#endif // SHR3D_WINDOW_WIN32
};

static std::vector<ClapPlugin> clapPlugins;

static void loadPluginInstance(ClapPlugin& clapPlugin)
{
  const i32 instance = i32(clapPlugin.pluginInstances.size());

  const clap_plugin_descriptor_t* desc = clapPlugin.pluginFactory->get_plugin_descriptor(clapPlugin.pluginFactory, 0);
  ASSERT(desc != nullptr);
  ASSERT(desc->clap_version.major >= 1);

  PluginInstance& pluginInstances = clapPlugin.pluginInstances.emplace_back();
  pluginInstances.plugin = clapPlugin.pluginFactory->create_plugin(clapPlugin.pluginFactory, &host, desc->id);

  const clap_plugin_t* plugin = pluginInstances.plugin;
  ASSERT(plugin != nullptr);

  if (!plugin->init(plugin))
    ASSERT(false);

  if (!plugin->activate(plugin, sampleRate(), blockSize(), blockSize()))
    ASSERT(false);

  if (!plugin->start_processing(plugin))
    ASSERT(false);
}

void SfxPluginClap::init(std::vector<std::u8string>& sfxNames)
{
  const std::vector<std::u8string> paths = String::split(Settings::pathClap, u8';');
  for (const std::u8string& path : paths)
  {
    const std::filesystem::path fsPath = std::filesystem::path(path);
    if (!std::filesystem::exists(fsPath))
      continue;

    for (const auto& file : std::filesystem::directory_iterator(fsPath))
    {
      if (file.path().extension() != std::filesystem::path(u8".clap"))
        continue;

#ifdef PLATFORM_WINDOWS
      void* hinstLib = LoadLibrary(file.path().wstring().c_str());
#else // PLATFORM_WINDOWS
      void* hinstLib = dlopen(file.path().string().c_str(), RTLD_LAZY | RTLD_LOCAL);
#endif // PLATFORM_WINDOWS

      if (hinstLib == nullptr)
        continue;

      const clap_plugin_entry* pluginEntry = reinterpret_cast<const struct clap_plugin_entry*>(dlsym((dlhandle)hinstLib, "clap_entry"));
      ASSERT(pluginEntry != nullptr);
      pluginEntry->init(file.path().string().c_str());

      ClapPlugin& clapPlugin = clapPlugins.emplace_back();
      clapPlugin.pluginFactory = static_cast<const clap_plugin_factory*>(pluginEntry->get_factory(CLAP_PLUGIN_FACTORY_ID));
      ASSERT(clapPlugin.pluginFactory != nullptr);

      //loadPluginInstance(clapPlugin);
      //const clap_plugin_t* plugin = clapPlugin.pluginInstances[0].plugin;

      const u32 plugin_count = clapPlugin.pluginFactory->get_plugin_count(clapPlugin.pluginFactory);
      //ASSERT(plugin_count == 1);

      const clap_plugin_descriptor_t* plugin_descriptor = clapPlugin.pluginFactory->get_plugin_descriptor(clapPlugin.pluginFactory, 0);
      ASSERT(plugin_descriptor != nullptr);

      sfxNames.emplace_back(reinterpret_cast<const char8_t*>(plugin_descriptor->name));

      //const clap_plugin_audio_ports* pluginAudioPorts = static_cast<const clap_plugin_audio_ports*>(plugin->get_extension(plugin, CLAP_EXT_AUDIO_PORTS));
      //const clap_plugin_gui* pluginGui = static_cast<const clap_plugin_gui*>(plugin->get_extension(plugin, CLAP_EXT_GUI));
      //const clap_plugin_latency* pluginLatency = static_cast<const clap_plugin_latency*>(plugin->get_extension(plugin, CLAP_EXT_LATENCY));
      //const clap_plugin_note_ports* pluginNotePorts = static_cast<const clap_plugin_note_ports*>(plugin->get_extension(plugin, CLAP_EXT_NOTE_PORTS));
      //const clap_plugin_params* pluginParams = static_cast<const clap_plugin_params*>(plugin->get_extension(plugin, CLAP_EXT_PARAMS));
      //const clap_plugin_posix_fd_support* pluginPosixFdSupport = static_cast<const clap_plugin_posix_fd_support*>(plugin->get_extension(plugin, CLAP_EXT_POSIX_FD_SUPPORT));
      //const clap_plugin_preset_load* pluginPresetLoad = static_cast<const clap_plugin_preset_load*>(plugin->get_extension(plugin, CLAP_EXT_PRESET_LOAD));
      //if (!pluginPresetLoad)
      //  pluginPresetLoad = static_cast<const clap_plugin_preset_load*>(plugin->get_extension(plugin, CLAP_EXT_PRESET_LOAD_COMPAT));
      //const clap_plugin_remote_controls* pluginRemoteControls = static_cast<const clap_plugin_remote_controls*>(plugin->get_extension(plugin, CLAP_EXT_REMOTE_CONTROLS));
      //if (!pluginRemoteControls)
      //  pluginRemoteControls = static_cast<const clap_plugin_remote_controls*>(plugin->get_extension(plugin, CLAP_EXT_REMOTE_CONTROLS_COMPAT));
      //const clap_plugin_render* pluginRender = static_cast<const clap_plugin_render*>(plugin->get_extension(plugin, CLAP_EXT_RENDER));
      //const clap_plugin_state* pluginState = static_cast<const clap_plugin_state*>(plugin->get_extension(plugin, CLAP_EXT_STATE));
      //const clap_plugin_tail* pluginTail = static_cast<const clap_plugin_tail*>(plugin->get_extension(plugin, CLAP_EXT_TAIL));
      //const clap_plugin_thread_pool* pluginThreadPool = static_cast<const clap_plugin_thread_pool*>(plugin->get_extension(plugin, CLAP_EXT_THREAD_POOL));
      //const clap_plugin_timer_support* pluginTimerSupport = static_cast<const clap_plugin_timer_support*>(plugin->get_extension(plugin, CLAP_EXT_TIMER_SUPPORT));
    }
  }
}

#ifdef SHR3D_SFX_PLUGIN_GUI_WINDOW
bool SfxPluginClap::hasWindow(i32 index)
{
  const clap_plugin_t* plugin = clapPlugins[index].pluginInstances[0].plugin;
  const clap_plugin_gui* pluginGui = static_cast<const clap_plugin_gui*>(plugin->get_extension(plugin, CLAP_EXT_GUI));

  if (pluginGui == nullptr)
    return false;

  return true;
}

void SfxPluginClap::openWindow(i32 index, i32 instance)
{
  ClapPlugin& clapPlugin = clapPlugins[index];
  const clap_plugin_t* plugin = clapPlugin.pluginInstances[instance].plugin;
  const clap_plugin_gui* pluginGui = static_cast<const clap_plugin_gui*>(plugin->get_extension(plugin, CLAP_EXT_GUI));

  ASSERT(pluginGui->is_api_supported(plugin, CLAP_WINDOW_API_OS, false));

  if (!pluginGui->create(plugin, CLAP_WINDOW_API_OS, false))
    ASSERT(false);

#ifdef SHR3D_WINDOW_SDL
  SDL_SysWMinfo wmInfo;
  SDL_VERSION(&wmInfo.version);
  SDL_GetWindowWMInfo(Global::window, &wmInfo);
#ifdef PLATFORM_WINDOWS
  HWND window = wmInfo.info.win.window;
  intptr_t display = 0;
#else // PLATFORM_WINDOWS
  void* window = (void*)wmInfo.info.x11.window;
  intptr_t display = (intptr_t)wmInfo.info.x11.display;
#endif // PLATFORM_WINDOWS
#endif // SHR3D_WINDOW_SDL
#ifdef SHR3D_WINDOW_WIN32
  HWND window = Global::window;
  intptr_t display = 0;
#endif // SHR3D_WINDOW_WIN32

  static const clap_window parentWindow
  {
    .api = CLAP_WINDOW_API_OS,
#ifdef PLATFORM_LINUX
    .x11 = (clap_xwnd)window,
#endif // PLATFORM_LINUX
#ifdef PLATFORM_WINDOWS
    .win32 = window
#endif // PLATFORM_WINDOWS
  };

  pluginGui->set_parent(plugin, &parentWindow);
  pluginGui->show(plugin);

#ifdef PLATFORM_WINDOWS
#ifdef SHR3D_WINDOW_SDL
  HWND childHwnd = GetWindow(window, GW_CHILD);
  clapPlugin.window = SDL_CreateWindowFrom((void*)childHwnd);
#else // SHR3D_WINDOW_SDL
  clapPlugin.window = GetWindow(window, GW_CHILD);
#endif // SHR3D_WINDOW_SDL
#else // PLATFORM_WINDOWS
  {
    ClapWindow wRoot, wParent, * pwChildren;
    unsigned int iChildren = 0;
    do
    {
      XQueryTree(wmInfo.info.x11.display, wmInfo.info.x11.window, &wRoot, &wParent, &pwChildren, &iChildren);
    } while (iChildren < 1); // this will deadlock when no window was opened.
    clapPlugin.window = SDL_CreateWindowFrom((void*)pwChildren[0]);
    XFree(pwChildren);
  }
#endif // PLATFORM_WINDOWS
}

Size SfxPluginClap::getWindowSize(i32 index, i32 instance)
{
  const clap_plugin_t* plugin = clapPlugins[index].pluginInstances[instance].plugin;
  const clap_plugin_gui* pluginGui = static_cast<const clap_plugin_gui*>(plugin->get_extension(plugin, CLAP_EXT_GUI));

  Size size;
  pluginGui->get_size(plugin, &size.width, &size.height);

  return size;
}

void SfxPluginClap::moveWindow(i32 index, i32 instance, i32 x, i32 y)
{
  ClapPlugin& clapPlugin = clapPlugins[index];
  const clap_plugin_t* plugin = clapPlugin.pluginInstances[instance].plugin;
  const clap_plugin_gui* pluginGui = static_cast<const clap_plugin_gui*>(plugin->get_extension(plugin, CLAP_EXT_GUI));

  u32 width;
  u32 height;
  pluginGui->get_size(plugin, &width, &height);

#ifdef SHR3D_WINDOW_SDL
  SDL_SetWindowPosition(clapPlugin.window, x, y);
#else // SHR3D_WINDOW_SDL
  MoveWindow(clapPlugin.window, x, y, width, height, TRUE);
#endif // SHR3D_WINDOW_SDL
}

void SfxPluginClap::closeWindow(i32 index, i32 instance)
{
  const clap_plugin_t* plugin = clapPlugins[index].pluginInstances[instance].plugin;
  const clap_plugin_gui* pluginGui = static_cast<const clap_plugin_gui*>(plugin->get_extension(plugin, CLAP_EXT_GUI));

  pluginGui->hide(plugin);
  pluginGui->destroy(plugin);
}
#endif // SHR3D_SFX_PLUGIN_GUI_WINDOW


i32 SfxPluginClap::numParams(i32 index, i32 instance)
{
  const clap_plugin_t* plugin = clapPlugins[index].pluginInstances[instance].plugin;
  const clap_plugin_params* pluginParams = static_cast<const clap_plugin_params*>(plugin->get_extension(plugin, CLAP_EXT_PARAMS));

  //plugin->stop_processing(plugin);

  const u32 paramCount = pluginParams->count(plugin);

  return paramCount;
}

void SfxPluginClap::getParameterProperties(i32 index, i32 instance, i32 param, SfxParameterProperties& parameterProperties)
{
  const clap_plugin_t* plugin = clapPlugins[index].pluginInstances[instance].plugin;
  const clap_plugin_params* pluginParams = static_cast<const clap_plugin_params*>(plugin->get_extension(plugin, CLAP_EXT_PARAMS));

  const u32 paramCount = pluginParams->count(plugin);

  clap_param_info_t paramInfo;
  pluginParams->get_info(plugin, param, &paramInfo);

  memcpy(parameterProperties.name, paramInfo.name, sizeof(parameterProperties.name));

  f64 value;
  pluginParams->get_value(plugin, paramInfo.id, &value);

  parameterProperties.label[0] = '\0'; // not used on clap

  pluginParams->value_to_text(plugin, paramInfo.id, value, parameterProperties.display, sizeof(parameterProperties.display));
}

f32 SfxPluginClap::getParameter(i32 index, i32 instance, i32 param)
{
  const clap_plugin_t* plugin = clapPlugins[index].pluginInstances[instance].plugin;
  const clap_plugin_params* pluginParams = static_cast<const clap_plugin_params*>(plugin->get_extension(plugin, CLAP_EXT_PARAMS));

  const u32 paramCount = pluginParams->count(plugin);

  clap_param_info_t paramInfo;
  pluginParams->get_info(plugin, param, &paramInfo);

  f64 value;
  pluginParams->get_value(plugin, paramInfo.id, &value);

  return value;
}

void SfxPluginClap::setParameter(i32 index, i32 instance, i32 param, f32 value)
{
  ClapPlugin& clapPlugin = clapPlugins[index];
  PluginInstance& pluginInstances = clapPlugin.pluginInstances[instance];
  const clap_plugin_t* plugin = pluginInstances.plugin;
  const clap_plugin_params* pluginParams = static_cast<const clap_plugin_params*>(plugin->get_extension(plugin, CLAP_EXT_PARAMS));

  const u32 paramCount = pluginParams->count(plugin);

  clap_param_info_t paramInfo;
  pluginParams->get_info(plugin, param, &paramInfo);

  f64 currentValue;
  pluginParams->get_value(plugin, paramInfo.id, &currentValue);

  if (abs(value - f32(currentValue)) > 0.01)
  {
    clap_event_param_value ev;
    ev.header.time = 0;
    ev.header.type = CLAP_EVENT_PARAM_VALUE;
    ev.header.space_id = CLAP_CORE_EVENT_SPACE_ID;
    ev.header.flags = 0;
    ev.header.size = sizeof(ev);
    ev.param_id = paramInfo.id;
    ev.cookie = nullptr;
    ev.port_index = 0;
    ev.key = -1;
    ev.channel = -1;
    ev.note_id = -1;
    ev.value = value;

    clapTryPush(pluginInstances.out_events, &ev.header);

    pluginInstances.sheduleEventFlush = true;
  }
}

void SfxPluginClap::processBlock(i32 index, i32 instance, f32** inBlock, f32** outBlock, i32 blockSize)
{
  ClapPlugin& clapPlugin = clapPlugins[index];
  PluginInstance& pluginInstances = clapPlugin.pluginInstances[instance];

  {
    const std::unique_lock<std::mutex> guard(loadSfxPluginClapInstanceMutex);
    while (clapPlugin.pluginInstances.size() <= instance)
      loadPluginInstance(clapPlugin);
  }

  pluginInstances.audio_inputs->data32 = inBlock;
  pluginInstances.audio_outputs->data32 = outBlock;
  pluginInstances.process->frames_count = u32(blockSize);
  //pluginInstances.process.audio_inputs = &pluginInstances.audioIn;
  //pluginInstances.process.audio_outputs = &pluginInstances.audioOut;
  //pluginInstances.process.in_events = &pluginInstances.inputEvents;
  //pluginInstances.process.out_events = &pluginInstances.outputEvents;

  const clap_plugin_t* plugin = pluginInstances.plugin;

  if (pluginInstances.sheduleEventFlush)
  {
    const clap_plugin_params* pluginParams = static_cast<const clap_plugin_params*>(plugin->get_extension(plugin, CLAP_EXT_PARAMS));
    plugin->stop_processing(plugin);
    pluginParams->flush(plugin, pluginInstances.in_events, pluginInstances.out_events);
    plugin->start_processing(plugin);
  }

  if (plugin->process(plugin, pluginInstances.process) != CLAP_PROCESS_CONTINUE)
  {
    //ASSERT(false);
  }

  pluginInstances.process->steady_time += blockSize;
}

std::u8string SfxPluginClap::saveParameters(i32 index, i32 instance)
{
  const clap_plugin_t* plugin = clapPlugins[index].pluginInstances[instance].plugin;
  const clap_plugin_params* pluginParams = static_cast<const clap_plugin_params*>(plugin->get_extension(plugin, CLAP_EXT_PARAMS));

  plugin->stop_processing(plugin);

  const u32 paramCount = pluginParams->count(plugin);

  std::u8string params;
  for (i32 i = 0; i < paramCount; ++i)
  {
    clap_param_info_t paramInfo;
    pluginParams->get_info(plugin, i, &paramInfo);

    char8_t substr[256];
    f64 value;
    pluginParams->get_value(plugin, paramInfo.id, &value);
    pluginParams->value_to_text(plugin, paramInfo.id, value, (char*)substr, sizeof(substr));
    params += std::u8string(substr) + u8',';
  }
  params.pop_back();

  return params;
}

void SfxPluginClap::loadParameters(i32 index, i32 instance, const std::u8string& parameters)
{
  ClapPlugin& clapPlugin = clapPlugins[index];

  {
    const std::unique_lock<std::mutex> guard(loadSfxPluginClapInstanceMutex);
    while (clapPlugin.pluginInstances.size() == instance)
      loadPluginInstance(clapPlugin);
  }

  PluginInstance& pluginInstances = clapPlugin.pluginInstances[instance];
  const clap_plugin_t* plugin = pluginInstances.plugin;
  const clap_plugin_params* pluginParams = static_cast<const clap_plugin_params*>(plugin->get_extension(plugin, CLAP_EXT_PARAMS));
  const u32 paramCount = pluginParams->count(plugin);

  u64 strLength = parameters.size();
  char8_t substr[256];
  u64 substrIndex = 0;
  u32 paramIndex = 0;
  for (u64 i = 0; i <= strLength; ++i) // i is out of bound at the end
  {
    if (parameters[i] == u8',' || i == strLength)
    {
      ASSERT(paramIndex < paramCount);

      clap_param_info_t paramInfo;
      pluginParams->get_info(plugin, paramIndex, &paramInfo);
      ++paramIndex;

      substr[i - substrIndex] = '\0';

      f64 value;
      pluginParams->text_to_value(plugin, paramInfo.id, (char*)substr, &value);
      
      clap_event_param_value ev;
      ev.header.time = 0;
      ev.header.type = CLAP_EVENT_PARAM_VALUE;
      ev.header.space_id = CLAP_CORE_EVENT_SPACE_ID;
      ev.header.flags = 0;
      ev.header.size = sizeof(ev);
      ev.param_id = paramInfo.id;
      ev.cookie = nullptr;
      ev.port_index = 0;
      ev.key = -1;
      ev.channel = -1;
      ev.note_id = -1;
      ev.value = value;

      clapTryPush(pluginInstances.out_events, &ev.header);

      substrIndex = i + 1;
    }
    substr[i - substrIndex] = parameters[i]; // carefull overrun at the end. It is not accessed so it should be fine.
  }

  pluginInstances.sheduleEventFlush = true;
}

#endif // SHR3D_SFX_PLUGIN_CLAP
