// SPDX-License-Identifier: Unlicense

#include "stats.h"

#include "file.h"
#include "global.h"
#include "ini.h"
#include "string_.h"

static std::map<std::u8string, std::map<std::u8string, std::u8string>> serialize(SongIndex songIndex)
{
  const std::vector<Manifest::Info>& manifestInfos = Global::songInfos[songIndex].manifestInfos;

  std::map<std::u8string, std::map<std::u8string, std::u8string>> serializedSaves;

  for (const auto& manifestInfo : manifestInfos)
  {
    std::map<std::u8string, std::u8string> serializedSave;

    if (Global::songProfile.contains(manifestInfo.persistentId))
    {
      const SongProfile& songProfile = Global::songProfile[manifestInfo.persistentId];

      if (songProfile.timesPlayed > 0)
      {
        serializedSave.insert({ u8"TimesPlayed", to_string(songProfile.timesPlayed)});
      }
      if (songProfile.lastPlayed > 0)
      {
        serializedSave.insert({ u8"LastPlayed", to_string(songProfile.lastPlayed) });
      }
      if (songProfile.score > 0.0f)
      {
        serializedSave.insert({ u8"Score", to_string(songProfile.score) });
      }
      if (songProfile.toneBank >= 0)
      {
        serializedSave.insert({ u8"ToneBank", to_string(songProfile.toneBank) });
      }
      for (i32 i = 0; i < songProfile.toneOverwrites.size(); ++i)
      {
        serializedSave.insert({ u8"ToneOverwrite" + to_string(i), songProfile.toneOverwrites[i] });
      }
    }

    if (serializedSave.size() >= 1)
      serializedSaves.insert({ manifestInfo.persistentId, serializedSave });
  }

  return serializedSaves;
}

static void loadSongProfile(const std::pair<std::u8string, std::map<std::u8string, std::u8string>>& entry)
{
  SongProfile& songProfile = Global::songProfile[entry.first];
  if (entry.second.contains(u8"LastPlayed"))
    songProfile.lastPlayed = atoi(reinterpret_cast<const char*>(entry.second.at(u8"LastPlayed").c_str()));
  if (entry.second.contains(u8"Score"))
    songProfile.score = f32(atof(reinterpret_cast<const char*>(entry.second.at(u8"Score").c_str())));
  if (entry.second.contains(u8"TimesPlayed"))
    songProfile.timesPlayed = atoi(reinterpret_cast<const char*>(entry.second.at(u8"TimesPlayed").c_str()));
  if (entry.second.contains(u8"ToneBank"))
    songProfile.toneBank = atoi(reinterpret_cast<const char*>(entry.second.at(u8"ToneBank").c_str()));
}

void Stats::init()
{
  if (!File::exists(Settings::pathStatsIni.c_str()))
    return;

  const std::vector<u8> statsIniData = File::read(Settings::pathStatsIni.c_str());
  const std::map<std::u8string, std::map<std::u8string, std::u8string>> serializedSaves = Ini::loadIniContent((const char8_t*)statsIniData.data(), statsIniData.size());

  for (const std::pair<std::u8string, std::map<std::u8string, std::u8string>>& pair : serializedSaves)
  {
    loadSongProfile(pair);
  }
}

void Stats::fini()
{
  std::map<std::u8string, std::map<std::u8string, std::u8string>> serializedSaves;

  for (SongIndex songIndex = 0; songIndex < SongIndex(Global::songInfos.size()); ++songIndex)
  {
    const std::map<std::u8string, std::map<std::u8string, std::u8string>> map = serialize(songIndex);
    serializedSaves.insert(map.begin(), map.end());
  }

  Ini::saveIniFile(Settings::pathStatsIni.c_str(), serializedSaves);
}
