// SPDX-License-Identifier: Unlicense

#include "tones.h"

#include "base64.h"
#include "file.h"
#include "ini.h"
#include "global.h"
#include "sfx.h"

#ifdef SHR3D_SFX
static void loadSfxTone(const std::map<std::u8string, std::u8string>& sfxTone)
{
  for (const auto& [key, value] : sfxTone)
  {
    ASSERT(key.size() >= 1);

    SfxToneIndex assignmentIndex;
    i32 slotIndex = -1;
    {
      const char* str = reinterpret_cast<const char*>(key.c_str());
      const i32 bankIndex = atoi(str); // atoi stops at ',' We are good.
      i32 toneIndex = -1;
      str += 1;
      while (str[0] != '\0')
      {
        if (str[0] == ',')
        {
          if (toneIndex < 0)
          {
            toneIndex = atoi(str + 1);
          }
          else
          {
            slotIndex = atoi(str + 1);
            break;
          }
        }
        str = str + 1;
      }
      ASSERT(toneIndex / Const::sfxToneTonesPerBank == 0);
      assignmentIndex = bankIndex * Const::sfxToneTonesPerBank + toneIndex;
    }

    if (slotIndex < 0) // value is bank name
    {
      Global::sfxToneNames[assignmentIndex] = value;
    }
    else // is sfx
    {
      if (slotIndex > ARRAY_SIZE(Global::effectChain) - 1)
        continue;

      const u64 comma0 = value.find(u8',');
      const u64 comma1 = value.find(u8',', comma0 + 1);

      const std::u8string sfxTypeStr = value.substr(0, comma0);
      const i32 indexBegin = Sfx::typeName2indexBegin(sfxTypeStr);
      const i32 indexEnd = Sfx::indexBegin2indexEnd(indexBegin);

      const std::u8string sfxName = value.substr(comma0 + 1, comma1 - comma0 - 1);

      for (i32 k = indexBegin; k <= indexEnd; ++k)
      {
        if (sfxName != Sfx::names[k])
          continue;

        ASSERT(value[comma1 + 1] == '1' || value[comma1 + 1] == '0');
        const bool isSfxMuted = value[comma1 + 1] == '0';

        Global::sfxTone[assignmentIndex][slotIndex] = isSfxMuted ? -k : k;
        const std::u8string parameters = value.substr(comma1 + 3);

#ifdef SHR3D_SFX_CORE_RAKARRACK
        if (k < to_underlying_(SfxCore::COUNT))
        {
          Global::sfxParameters[assignmentIndex][slotIndex] = parameters;
        }
        else
#endif // SHR3D_SFX_CORE_RAKARRACK
        {
          const std::vector<u8> decodedData = Base64::decode(parameters.data(), parameters.size());
          const std::u8string decodedParameters(reinterpret_cast<const char8_t*>(decodedData.data()), decodedData.size());

          Global::sfxParameters[assignmentIndex][slotIndex] = decodedParameters;
        }
        break;
      }
    }
  }
}
#endif // SHR3D_SFX

void Tones::init()
{
  std::map<std::u8string, std::map<std::u8string, std::u8string>> serializedSaves;

  if (File::exists(Settings::pathTonesIni.c_str()))
  {
    const std::vector<u8> tonesIniData = File::read(Settings::pathTonesIni.c_str());
    serializedSaves = Ini::loadIniContent(reinterpret_cast<const char8_t*>(tonesIniData.data()), tonesIniData.size());
  }
#ifdef SHR3D_SFX
  else
  {
#ifdef SHR3D_SFX_CORE_RAKARRACK_PRESETS
    serializedSaves = { { u8"Tones", {
    { u8"0010G",   u8"Rodent Airlines" },
    { u8"0010G00", u8"Expander,1,-55,37,24,46,8050,76,1" },
    { u8"0010G01", u8"StompBox,1,32,-44,0,0,110,2" },
    { u8"0010G02", u8"Echotron,1,63,28,39,8,0,10,0,64,3,32,4,64,1,1,0,1" },
    { u8"0010G03", u8"Reverbtron,1,64,17,1,400,0,0,34,6,1,0,0,64,1,0,2524,0" },
    { u8"0011G",   u8"Ambient Flange" },
    { u8"0011G00", u8"Expander,1,-55,37,24,46,2091,20,1" },
    { u8"0011G01", u8"Sustainer,1,66,37" },
    { u8"0011G02", u8"Echotron,1,63,20,20,8,0,61,0,64,3,32,4,64,1,1,1,1" },
    { u8"0011G03", u8"Reverbtron,1,64,0,1,400,0,0,34,39,1,0,0,64,1,0,2524,76" },
    { u8"0012G",   u8"Cathedral Swell" },
    { u8"0012G00", u8"Expander,1,-32,20,1806,118,6673,20,46" },
    { u8"0012G01", u8"Sustainer,1,99,29" },
    { u8"0012G02", u8"Reverbtron,1,103,0,1,400,0,30,0,40,9,0,0,64,0,0,20000,0" },
    { u8"0013G",   u8"Go with Him" },
    { u8"0013G00", u8"Ring,1,-64,0,-64,127,44,392,0,74,18,48,17,81,1" },
    { u8"0013G01", u8"Synthfilter,1,54,59,180,0,0,91,67,-40,6,0,0,57,-17,77,176,53" },
    { u8"0013G02", u8"Echoverse,1,119,64,90,52,37,64,64,0,1,114" },
    { u8"0013G03", u8"Pan,1,37,64,180,0,0,127,74,1,1" },
    { u8"0013G04", u8"Reverb,1,69,64,30,24,0,3090,5,83,1,83" },
    { u8"0014G",   u8"All is Lost" },
    { u8"0014G00", u8"EQ,1,64,71,75,71,64,62,68,73,64,32,56,64" },
    { u8"0014G01", u8"Echoverse,1,64,64,132,64,64,29,10,10,3,64" },
    { u8"0014G02", u8"Arpie,1,39,64,132,64,100,72,62,127,6,3,3" },
    { u8"0014G03", u8"Chorus,1,64,64,33,0,0,68,40,45,33,119,0,0" },
    { u8"0014G04", u8"Phaser,1,64,64,11,0,0,64,40,64,1,0,0,20" },
    { u8"0014G05", u8"Reverb,1,80,64,63,24,0,85,5,83,1,64" },
    { u8"0015G",   u8"Tight Rock" },
    { u8"0015G00", u8"Expander,1,-55,18,50,50,8050,76,0" },
    { u8"0015G01", u8"Distorsion,1,0,64,64,127,76,26,1,2982,645,0,1,0,0" },
    { u8"0015G02", u8"Overdrive,1,0,64,35,9,44,26,0,3327,21,0,0,0,0" },
    { u8"0015G03", u8"Reverbtron,1,64,0,1,400,0,0,99,25,0,0,0,64,0,0,0,0" },
    { u8"0015G04", u8"Parametric EQ,1,127,76,12,931,53,64,8111,64,64,64" },
    { u8"0016G",   u8"Summer at the Pool" },
    { u8"0016G00", u8"Expander,1,-64,22,50,50,8050,76,0" },
    { u8"0016G01", u8"Sustainer,1,36,91" },
    { u8"0016G02", u8"Echotron,1,55,72,37,4,0,84,3,47,0,96,-6,64,1,0,0,1" },
    { u8"0017G",   u8"Summer at the Pit" },
    { u8"0017G00", u8"Expander,1,-64,22,50,50,8050,76,0" },
    { u8"0017G01", u8"StompBox,1,42,-20,0,0,70,2" },
    { u8"0017G02", u8"Echotron,1,51,72,37,4,0,73,3,47,0,96,-6,64,1,0,0,1" },
    { u8"0018G",   u8"Summer at the Beach" },
    { u8"0018G00", u8"Expander,1,-64,22,50,50,8050,76,0" },
    { u8"0018G01", u8"Sustainer,1,36,91" },
    { u8"0018G02", u8"Analog Phaser,1,64,20,25,50,1,64,93,40,6,48,0,58,1" },
    { u8"0018G03", u8"Echotron,1,64,45,34,4,0,76,3,41,0,96,-13,64,1,1,0,1" },
    { u8"0019G",   u8"Suction" },
    { u8"0019G00", u8"Expander,1,-64,22,50,50,1200,70,0" },
    { u8"0019G01", u8"Sustainer,1,75,73" },
    { u8"0019G02", u8"Echotron,1,0,47,28,8,0,92,0,64,3,32,0,64,1,1,0,1" },
    { u8"0019G03", u8"Reverbtron,1,104,0,1,400,0,0,60,15,2,0,0,64,0,0,20000,20" },
    { u8"0020G",   u8"Suction Flange +dist" },
    { u8"0020G00", u8"Expander,1,-64,22,50,50,8050,76,0" },
    { u8"0020G01", u8"Sustainer,1,75,73" },
    { u8"0020G02", u8"StompBox,1,37,64,0,0,127,4" },
    { u8"0020G03", u8"Echotron,1,64,36,93,8,0,81,0,64,3,32,0,64,1,0,0,1" },
    { u8"0020G04", u8"Reverbtron,1,104,0,1,155,0,0,60,21,2,0,0,64,0,0,20000,37" },
    { u8"0021G",   u8"Succulent Clean" },
    { u8"0021G00", u8"Expander,1,-57,19,10,10,8050,76,72" },
    { u8"0021G01", u8"MuTroMojo,1,28,8,35,0,0,64,50,-6,53,58,28,-16,32,2,2600,300,1,0" },
    { u8"0021G02", u8"DualFlange,1,-40,0,0,35,9,67,12,4700,1,1,160,75,0,60" },
    { u8"0022G",   u8"EboWeboAmiYeebo" },
    { u8"0022G00", u8"Expander,1,-64,22,50,50,8050,483,0" },
    { u8"0022G01", u8"Sustainer,1,58,54" },
    { u8"0022G02", u8"Echoverse,1,64,64,101,48,33,77,0,0,1,67" },
    { u8"0022G03", u8"Sequence,1,30,127,30,49,80,40,110,80,44,101,95,1,0,2,2" },
    { u8"0023G",   u8"Rock Amp" },
    { u8"0023G00", u8"Expander,1,-55,37,24,46,8050,76,1" },
    { u8"0023G01", u8"Sustainer,1,36,51" },
    { u8"0023G02", u8"StompBox,1,55,22,15,0,21,0" },
    { u8"0023G03", u8"Convolotron,1,0,64,0,76,0,0,0,72,2,0,-61" },
    { u8"0023G04", u8"Reverbtron,1,108,0,0,181,0,0,0,0,3,0,0,64,1,1,20000,21" },
    { u8"0024G",   u8"Overdrive" },
    { u8"0024G00", u8"StompBox,1,127,19,-23,0,5,0" },
    { u8"0024G01", u8"Reverbtron,1,108,0,0,155,0,0,0,0,3,0,0,64,1,1,20000,12" },
    { u8"0025G",   u8"Metal Amp" },
    { u8"0025G00", u8"Expander,1,-51,50,59,46,8050,76,1" },
    { u8"0025G01", u8"StompBox,1,48,0,-32,-36,53,6" },
    { u8"0025G02", u8"Convolotron,1,0,64,0,76,0,0,0,88,0,0,0" },
    { u8"0025G03", u8"Reverbtron,1,64,0,0,327,0,0,0,0,7,0,0,64,0,0,1585,12" },
    { u8"0026G",   u8"Hard Rock Amp" },
    { u8"0026G00", u8"Expander,1,-55,37,24,46,8050,76,1" },
    { u8"0026G01", u8"Overdrive,1,0,64,35,79,48,26,1,5406,708,0,1,0,0" },
    { u8"0026G02", u8"StompBox,1,120,-3,-21,28,51,0" },
    { u8"0026G03", u8"Convolotron,1,0,64,0,134,0,0,0,51,3,0,0" },
    { u8"0026G04", u8"Reverbtron,1,95,0,0,155,0,0,99,6,0,0,0,64,0,0,20000,21" },
    { u8"0027G",   u8"Edgy" },
    { u8"0027G00", u8"Expander,1,-55,18,50,50,8050,20,0" },
    { u8"0027G01", u8"StompBox,1,78,10,-2,5,37,0" },
    { u8"0027G02", u8"Reverbtron,1,64,0,1,110,0,0,48,20,4,0,0,64,0,0,0,0" },
    { u8"0027G03", u8"Looper,1,64,1,0,0,1,0,70,1,0,1,70,1,0,0" },
    { u8"0028G",   u8"Dark Alley" },
    { u8"0028G00", u8"Expander,1,-64,22,50,50,8050,76,0" },
    { u8"0028G01", u8"Shifter,1,0,64,64,200,200,-9,1,0,0,0" },
    { u8"0028G02", u8"Sustainer,1,79,83" },
    { u8"0028G03", u8"DualFlange,1,-32,0,64,60,10,100,20,16000,0,0,16,90,4,0" },
    { u8"0028G04", u8"Echoverse,1,64,64,90,64,64,64,64,0,1,64" },
    { u8"0029G",   u8"D Flange" },
    { u8"0029G00", u8"Expander,1,-64,22,50,50,8050,76,0" },
    { u8"0029G01", u8"DualFlange,1,-32,0,0,110,800,10,-27,16000,1,0,40,64,1,10" },
    { u8"0029G02", u8"Sustainer,1,45,107" },
    { u8"0029G03", u8"Echoverse,1,68,64,81,64,33,39,0,0,1,67" },
    { u8"0029G04", u8"Reverb,1,121,64,62,24,0,3134,102,75,1,49" },
    { u8"0030G",   u8"Clean Swirl" },
    { u8"0030G00", u8"Expander,1,-64,22,50,50,8050,76,0" },
    { u8"0030G01", u8"Sustainer,1,39,91" },
    { u8"0030G02", u8"Analog Phaser,1,64,20,32,50,1,64,93,40,6,48,0,58,1" },
    { u8"0030G03", u8"Echoverse,1,111,64,90,64,64,18,23,0,1,82" },
    { u8"0030G04", u8"Reverb,1,116,64,62,24,0,3134,102,75,1,49" },
    { u8"0030G05", u8"Looper,1,64,0,1,0,1,0,64,1,0,1,64,1,0,0" },
    { u8"0031G",   u8"Noise Synth" },
    { u8"0031G00", u8"Expander,1,-49,28,50,50,8050,76,27" },
    { u8"0031G01", u8"StompBox,1,127,-31,0,0,0,7" },
    { u8"0031G02", u8"Synthfilter,1,0,5,75,0,1,64,0,-32,6,0,0,0,11,439,122,0" },
    { u8"0031G03", u8"Echo,1,8,64,575,65,55,105,77,0,0" },
    { u8"0031G04", u8"MuTroMojo,1,28,8,35,0,0,64,50,-6,53,58,28,-16,32,4,2600,300,1,0" },
    { u8"0031G05", u8"DualFlange,1,-15,0,0,42,12,50,-10,1500,0,0,120,0,0,20" },
    { u8"0031G06", u8"Parametric EQ,1,72,64,64,1077,76,64,8111,64,64,104" },
    { u8"0032G",   u8"Acoustic Sparkle" },
    { u8"0032G00", u8"Compressor,1,-24,4,0,45,135,1,84,0,1" },
    { u8"0032G01", u8"Distorsion,1,81,64,64,90,81,26,0,9586,15641,1,1,0,0" },
    { u8"0032G02", u8"Reverb,1,88,64,53,20,0,25040,2417,71,1,64" },
    { u8"0032G03", u8"Parametric EQ,1,256,29,64,1932,39,82,11766,66,72,64" },
    { u8"0033G",   u8"Extreme" },
    { u8"0033G00", u8"EQ,1,17,31,32,36,42,34,42,56,59,72,58,127" },
    { u8"0033G01", u8"Distorsion,1,0,64,64,19,59,1,0,25040,21,1,0,0,0" },
    { u8"0033G02", u8"Echo,1,49,64,534,62,32,64,0,1,0" },
    { u8"0033G03", u8"Chorus,1,33,64,106,36,0,50,18,34,11,0,0,1" },
    { u8"0033G04", u8"Phaser,1,41,64,4,43,0,24,33,98,8,0,0,110" },
    { u8"0033G05", u8"Flanger,1,46,64,2,34,1,95,39,62,18,0,0,1" },
    { u8"0034G",   u8"Heaven Garden" },
    { u8"0034G00", u8"Distorsion,1,80,64,0,24,8,0,0,7678,56,0,0,0,0" },
    { u8"0034G01", u8"WahWah,1,112,64,140,0,0,64,70,64,0,60" },
    { u8"0034G02", u8"Harmonizer,1,20,48,64,24,600,0,9,0,64,64,0" },
    { u8"0034G03", u8"Musical Delay,1,64,0,3,7,125,49,0,127,6,89,90,96,44" },
    { u8"0034G04", u8"Pan,1,0,64,26,0,0,0,32,0,1" },
    { u8"0034G05", u8"Reverb,1,80,64,40,24,0,4002,27,83,1,64" },
    { u8"0035G",   u8"Old Pat" },
    { u8"0035G00", u8"Distorsion,1,0,64,64,7,57,6,0,692,21,0,0,1,0" },
    { u8"0035G01", u8"Echo,1,106,64,534,62,32,70,0,1,0" },
    { u8"0035G02", u8"Chorus,1,89,64,7,0,0,42,97,95,90,127,0,0" },
    { u8"0035G03", u8"Cabinet,1,3,64" },
    { u8"0036G",   u8"Metal Tone" },
    { u8"0036G00", u8"EQ,1,25,35,58,70,82,104,67,30,14,0,83,32" },
    { u8"0036G01", u8"Overdrive,1,51,64,35,21,48,14,0,824,21,0,0,1,0" },
    { u8"0036G02", u8"Distorsion,1,0,64,0,72,60,15,0,9586,21,0,0,1,0" },
    { u8"0036G03", u8"Parametric EQ,1,159,78,0,569,33,25,4767,57,9,86" },
    { u8"0037G",   u8"Dist-Duet-A7 *" },
    { u8"0037G00", u8"Distorsion,1,0,58,0,65,0,0,0,25040,209,0,0,0,0" },
    { u8"0037G01", u8"Harmonizer,1,74,104,64,16,600,1,9,19,64,64,0" },
    { u8"0037G02", u8"WahWah,1,116,64,39,0,0,50,54,54,0,60" },
    { u8"0037G03", u8"Musical Delay,1,113,0,2,7,0,59,0,127,4,59,60,75,75" },
    { u8"0037G04", u8"Cabinet,1,5,34" },
    { u8"0038G",   u8"Bass" },
    { u8"0038G00", u8"EQ,1,88,76,70,68,72,64,64,64,64,64,52,49" },
    { u8"0038G01", u8"Harmonizer,1,0,64,90,0,52,0,0,0,70,49,0" },
    { u8"0038G02", u8"Parametric EQ,1,55,70,61,870,64,55,8111,64,64,72" },
    { u8"0039G",   u8"Ballada Duet Gm *" },
    { u8"0039G00", u8"Distorsion,1,40,64,127,69,0,0,0,2837,408,0,1,1,0" },
    { u8"0039G01", u8"Harmonizer,1,64,0,64,8,600,1,7,10,64,64,0" },
    { u8"0039G02", u8"Chorus,1,97,64,19,0,0,98,56,90,64,0,0,0" },
    { u8"0039G03", u8"Echo,1,71,64,566,64,23,59,90,1,0" },
    { u8"0039G04", u8"Pan,1,64,64,26,0,0,0,27,0,1" },
    { u8"0039G05", u8"NoiseGate,1,-20,-23,1,132,6703,76,334" },
    { u8"0040G",   u8"True Heavy-3" },
    { u8"0040G00", u8"Distorsion,1,0,64,0,109,32,13,0,1157,37,1,1,1,0" },
    { u8"0040G01", u8"Pan,1,53,64,76,0,5,127,86,1,1" },
    { u8"0040G02", u8"Reverb,1,102,64,40,24,0,4002,27,83,1,64" },
    { u8"0040G03", u8"Cabinet,1,0,84" },
    { u8"0041G",   u8"Angel's Chorus" },
    { u8"0041G00", u8"WahWah,1,64,64,26,0,0,96,64,0,0,60" },
    { u8"0041G01", u8"Compressor,1,-26,4,-25,20,120,1,10,0,1" },
    { u8"0041G02", u8"Overdrive,1,66,64,35,29,34,1,0,25040,21,0,0,1,0" },
    { u8"0041G03", u8"Echo,1,67,64,347,64,30,59,0,1,64" },
    { u8"0041G04", u8"Chorus,1,64,64,19,0,0,98,56,90,64,19,0,0" },
    { u8"0041G05", u8"Reverb,1,80,64,69,35,0,25040,21,71,0,64" },
    { u8"0041G06", u8"Analog Phaser,1,64,20,14,5,1,64,110,40,6,10,0,70,1" },
    { u8"0042G",   u8"Funk Wah" },
    { u8"0042G00", u8"WahWah,1,0,64,1,0,0,64,0,109,0,0" },
    { u8"0042G01", u8"Compressor,1,-30,8,-22,20,63,1,30,1,1" },
    { u8"0042G02", u8"Parametric EQ,1,159,60,37,1077,90,55,4767,62,14,64" },
    { u8"0042G03", u8"Reverb,1,113,64,35,9,0,1785,27,106,1,64" },
    { u8"0043G",   u8"Chorus Riff" },
    { u8"0043G00", u8"EQ,1,45,38,28,43,54,71,58,76,26,0,61,68" },
    { u8"0043G01", u8"Overdrive,1,18,64,35,58,48,0,0,1785,21,0,0,1,0" },
    { u8"0043G02", u8"Chorus,1,57,58,9,0,0,69,82,35,56,127,0,0" },
    { u8"0043G03", u8"Parametric EQ,1,29,69,23,1077,42,64,8111,85,64,64" },
    { u8"0044G",   u8"Big Stack" },
    { u8"0044G00", u8"Distorsion,1,0,64,61,42,58,14,0,3293,804,0,1,1,16" },
    { u8"0044G01", u8"Overdrive,1,21,64,35,23,51,15,0,1292,22,0,0,1,0" },
    { u8"0044G02", u8"Reverb,1,115,64,63,24,0,4002,27,83,1,64" },
    { u8"0044G03", u8"Cabinet,1,7,64" },
    { u8"0045G",   u8"Comphaser" },
    { u8"0045G00", u8"Compressor,1,-55,5,-21,19,23,1,0,0,1" },
    { u8"0045G01", u8"Analog Phaser,1,64,20,32,50,1,64,93,40,6,48,0,58,1" },
    { u8"0046G",   u8"Jet Flanger" },
    { u8"0046G00", u8"EQ,1,18,40,53,58,68,66,61,43,26,0,89,68" },
    { u8"0046G01", u8"Distorsion,1,2,64,61,79,64,0,0,334,335,0,1,0,0" },
    { u8"0046G02", u8"Echo,1,0,64,254,64,64,18,0,1,0" },
    { u8"0046G03", u8"Flanger,1,23,64,8,0,0,60,23,3,69,0,0,0" },
    { u8"0046G04", u8"Reverb,1,99,64,35,18,0,4002,27,83,1,64" },
    { u8"0046G05", u8"Parametric EQ,1,128,70,89,825,55,40,7691,57,47,64" },
    { u8"0047G",   u8"Nostalgia" },
    { u8"0047G00", u8"EQ,1,37,55,62,56,69,72,65,56,49,44,64,64" },
    { u8"0047G01", u8"Distorsion,1,14,64,64,9,51,17,0,7678,94,0,1,1,0" },
    { u8"0047G02", u8"Pan,1,30,64,46,0,4,64,0,1,1" },
    { u8"0047G03", u8"Reverb,1,117,64,51,24,0,3293,21,83,1,32" },
    { u8"0048G",   u8"Panalog Phaser" },
    { u8"0048G00", u8"Compressor,1,-30,2,0,20,120,1,0,0,0" },
    { u8"0048G01", u8"Overdrive,1,60,64,35,8,102,14,0,2316,494,0,1,0,0" },
    { u8"0048G02", u8"Reverb,1,97,64,40,18,0,4002,27,83,1,47" },
    { u8"0048G03", u8"Pan,1,0,53,61,14,0,127,44,1,1" },
    { u8"0048G04", u8"Analog Phaser,1,64,20,61,0,1,74,69,30,6,60,1,64,1" },
    { u8"0048G05", u8"Chorus,1,0,0,1,0,0,0,0,0,0,0,0,0" },
    { u8"0048G06", u8"Parametric EQ,1,72,68,64,1077,64,64,6916,25,12,64" },
    { u8"0049G",   u8"Space Invasion" },
    { u8"0049G00", u8"StompBox,1,48,-8,-36,57,58,7" },
    { u8"0049G01", u8"AlienWah,1,7,64,57,0,0,96,102,118,31,0,36" },
    { u8"0049G02", u8"Reverbtron,1,79,0,1,400,0,0,39,44,1,0,-31,64,1,1,20000,12" },
    { u8"0050G",   u8"Clean-A-Duet *" },
    { u8"0050G00", u8"EQ,1,64,80,72,64,64,64,64,60,59,47,64,56" },
    { u8"0050G01", u8"Harmonizer,1,46,27,64,9,600,1,9,0,64,64,0" },
    { u8"0050G02", u8"Reverb,1,109,64,45,0,0,16683,435,64,1,85" },
    { u8"0050G03", u8"Cabinet,1,5,71" },
    { u8"0051G",   u8"Tricky Dick" },
    { u8"0051G00", u8"EQ,1,38,47,58,65,76,86,81,80,75,68,63,110" },
    { u8"0051G01", u8"Chorus,1,64,64,398,0,0,90,40,85,47,119,0,0" },
    { u8"0051G02", u8"Flanger,1,64,64,39,0,0,60,23,3,62,0,0,0" },
    { u8"0051G03", u8"Reverb,1,111,64,10,26,0,1223,21,77,1,45" },
    { u8"0052G",   u8"Dist-WahWah-1" },
    { u8"0052G00", u8"EQ,1,72,63,59,65,62,52,64,64,64,57,64,42" },
    { u8"0052G01", u8"Valve,1,0,64,64,127,48,0,5841,61,1,0,69,1,84" },
    { u8"0052G02", u8"WahWah,1,84,64,7,0,4,28,55,113,1,60" },
    { u8"0052G03", u8"Musical Delay,1,80,0,2,7,0,59,0,127,4,59,60,75,75" },
    { u8"0052G04", u8"Cabinet,1,1,70" },
    { u8"0053G",   u8"Rolling Echo" },
    { u8"0053G00", u8"Compressor,1,-10,2,0,20,120,1,0,0,1" },
    { u8"0053G01", u8"Distorsion,1,0,64,64,7,55,1,0,692,21,0,0,1,0" },
    { u8"0053G02", u8"Echo,1,0,64,1220,62,32,103,111,1,0" },
    { u8"0053G03", u8"Chorus,1,89,64,7,0,0,42,97,95,90,127,0,0" },
    { u8"0053G04", u8"Reverb,1,110,64,32,0,0,7340,21,64,1,71" },
    { u8"0054G",   u8"Gary Mooore" },
    { u8"0054G00", u8"EQ,1,18,40,53,58,68,66,61,45,26,0,89,68" },
    { u8"0054G01", u8"Overdrive,1,0,64,35,110,23,13,0,2982,21,0,0,0,0" },
    { u8"0054G02", u8"Echo,1,0,64,488,64,64,35,0,1,0" },
    { u8"0054G03", u8"Parametric EQ,1,55,28,72,704,65,104,5896,52,64,92" },
    { u8"0055G",   u8"Satriani" },
    { u8"0055G00", u8"EQ,1,18,40,53,58,68,71,64,53,26,0,40,74" },
    { u8"0055G01", u8"Overdrive,1,18,64,35,72,48,0,0,1364,21,0,0,0,0" },
    { u8"0055G02", u8"Echo,1,0,64,254,64,64,18,0,1,0" },
    { u8"0055G03", u8"Phaser,1,39,59,23,0,0,16,91,64,8,0,0,40" },
    { u8"0056G",   u8"SubOctave" },
    { u8"0056G00", u8"Distorsion,1,2,64,61,36,51,0,0,1364,76,0,1,1,47" },
    { u8"0056G01", u8"Reverb,1,80,64,27,12,0,4002,27,83,1,64" },
    { u8"0056G02", u8"Cabinet,1,0,76" },
    { u8"0057G",   u8"Lil Dirty" },
    { u8"0057G00", u8"Distorsion,1,7,64,0,12,65,14,0,2982,559,0,1,1,0" },
    { u8"0057G01", u8"EQ,1,64,77,74,64,57,59,56,50,47,44,79,61" },
    { u8"0058G",   u8"Quasi Acoustic" },
    { u8"0058G00", u8"EQ,1,66,73,63,45,33,45,56,69,80,26,56,64" },
    { u8"0058G01", u8"Chorus,1,85,64,14,0,0,90,66,48,64,119,0,0" },
    { u8"0058G02", u8"Reverb,1,118,64,37,7,13,5078,21,75,1,51" },
    { u8"0058G03", u8"Parametric EQ,1,76,77,69,1404,74,43,6218,73,32,68" },
    { u8"0059G",   u8"Satriani 2" },
    { u8"0059G00", u8"EQ,1,18,40,53,58,68,66,61,53,26,0,64,68" },
    { u8"0059G01", u8"Overdrive,1,12,64,35,65,48,0,0,1441,21,0,0,0,0" },
    { u8"0059G02", u8"Echo,1,0,64,254,64,64,18,0,1,0" },
    { u8"0059G03", u8"Phaser,1,39,59,23,0,0,16,91,64,8,0,0,40" },
    { u8"0059G04", u8"Reverb,1,82,64,13,18,0,4002,27,83,0,64" },
    { u8"0059G05", u8"Parametric EQ,1,128,70,89,1932,75,60,11766,72,35,80" },
    { u8"0060G",   u8"Classic Chorus" },
    { u8"0060G00", u8"EQ,1,18,40,53,58,51,71,82,69,49,0,88,68" },
    { u8"0060G01", u8"Chorus,1,33,63,43,0,0,73,49,51,64,127,0,0" },
    { u8"0060G02", u8"Reverb,1,99,64,35,18,0,4002,27,83,1,64" },
    { u8"0060G03", u8"Parametric EQ,1,128,70,89,825,55,40,7691,70,47,64" },
    { u8"0061G",   u8"Trash Chorus" },
    { u8"0061G00", u8"EQ,1,18,40,53,58,68,66,61,43,26,0,89,68" },
    { u8"0061G01", u8"Distorsion,1,2,64,61,79,36,0,0,334,335,0,1,0,0" },
    { u8"0061G02", u8"Echo,1,0,64,254,64,64,18,0,1,0" },
    { u8"0061G03", u8"Chorus,1,33,63,1,0,0,73,120,86,64,127,0,0" },
    { u8"0061G04", u8"Reverb,1,99,64,35,18,0,4002,27,83,1,64" },
    { u8"0061G05", u8"Parametric EQ,1,128,70,89,825,58,40,7691,57,47,96" },
    { u8"0062G",   u8"Ballada Solo" },
    { u8"0062G00", u8"EQ,1,45,38,28,43,54,59,46,21,7,0,56,68" },
    { u8"0062G01", u8"Distorsion,1,40,64,61,127,48,0,0,2837,76,0,1,0,0" },
    { u8"0062G02", u8"Echo,1,48,64,566,64,23,59,90,1,0" },
    { u8"0063G",   u8"Spandex" },
    { u8"0063G00", u8"Distorsion,1,18,64,0,85,23,14,0,1785,494,0,1,1,0" },
    { u8"0063G01", u8"Overdrive,1,25,64,35,53,58,17,0,1292,35,0,0,1,0" },
    { u8"0063G02", u8"Echo,1,88,64,566,64,30,59,0,1,0" },
    { u8"0063G03", u8"Reverb,1,47,64,21,10,0,652,21,93,1,24" },
    { u8"0064G",   u8"CrunchPhaser" },
    { u8"0064G00", u8"Distorsion,1,0,64,0,87,37,14,0,2564,157,0,1,1,0" },
    { u8"0064G01", u8"Analog Phaser,1,64,20,14,0,1,64,110,40,4,10,0,64,1" },
    { u8"0065G",   u8"Joe" },
    { u8"0065G00", u8"EQ,1,23,35,61,73,84,82,69,54,41,18,56,64" },
    { u8"0065G01", u8"Reverb,1,118,64,37,7,13,5078,21,75,1,51" },
    { u8"0065G02", u8"Parametric EQ,1,84,77,77,704,64,80,7293,58,77,72" },
    { u8"0066G",   u8"Falling in Fantasy" },
    { u8"0066G00", u8"Distorsion,1,69,64,64,58,67,26,0,9586,10441,1,1,0,0" },
    { u8"0066G01", u8"Analog Phaser,1,64,20,5,0,3,127,127,28,4,10,0,72,1" },
    { u8"0066G02", u8"Echo,1,55,64,1143,65,55,35,21,1,1" },
    { u8"0066G03", u8"Reverb,1,103,64,62,24,0,3134,102,75,1,49" },
    { u8"0066G04", u8"Parametric EQ,1,256,29,64,1932,39,82,11766,66,72,72" },
    { u8"0067G",   u8"Underwater I" },
    { u8"0067G00", u8"Distorsion,1,14,64,64,23,31,2,0,25040,21,0,0,1,0" },
    { u8"0067G01", u8"Phaser,1,9,64,31,63,1,117,71,8,2,0,0,40" },
    { u8"0067G02", u8"Echo,1,64,64,472,65,55,68,0,1,0" },
    { u8"0067G03", u8"Reverb,1,99,64,62,24,0,1441,21,83,1,49" },
    { u8"0067G04", u8"Parametric EQ,1,72,73,45,1077,64,64,8111,69,38,64" },
    { u8"0068G",   u8"Slap Echo" },
    { u8"0068G00", u8"EQ,1,18,40,53,58,51,71,82,69,49,0,89,68" },
    { u8"0068G01", u8"Echo,1,64,64,254,64,64,18,0,1,0" },
    { u8"0068G02", u8"Parametric EQ,1,128,75,89,968,23,53,9513,82,69,72" },
    { u8"0069G",   u8"Aural Phaser" },
    { u8"0069G00", u8"EQ,1,18,40,53,58,51,71,82,69,49,0,80,68" },
    { u8"0069G01", u8"Phaser,1,39,59,23,0,0,16,91,64,8,0,0,40" },
    { u8"0069G02", u8"Reverb,1,99,64,35,18,0,4002,27,83,1,64" },
    { u8"0069G03", u8"Parametric EQ,1,128,70,89,825,55,40,7691,70,47,64" },
    { u8"0070G",   u8"Nasty" },
    { u8"0070G00", u8"Distorsion,1,39,64,0,65,35,14,0,428,31,0,1,1,65" },
    { u8"0070G01", u8"Overdrive,1,35,64,35,65,35,11,0,1785,1203,0,1,1,0" },
    { u8"0070G02", u8"Echo,1,88,64,566,64,30,59,0,1,0" },
    { u8"0070G03", u8"Analog Phaser,1,64,20,14,10,0,64,110,80,7,10,1,45,1" },
    { u8"0070G04", u8"Reverb,1,47,64,21,10,0,652,21,93,1,24" },
    { u8"0071G",   u8"Aymimadre" },
    { u8"0071G00", u8"Parametric EQ,1,256,59,64,1932,76,82,11766,66,72,88" },
    { u8"0071G01", u8"Overdrive,1,0,64,35,127,23,6,0,314,494,0,1,1,0" },
    { u8"0071G02", u8"Chorus,1,62,64,1,0,1,18,120,85,64,119,0,0" },
    { u8"0071G03", u8"Echo,1,56,64,51,65,55,74,21,1,0" },
    { u8"0072G",   u8"Clean-2" },
    { u8"0072G00", u8"Parametric EQ,1,256,59,64,1932,76,82,11766,66,72,60" },
    { u8"0072G01", u8"Flanger,1,112,64,7,85,0,41,77,0,11,17,0,1" },
    { u8"0072G02", u8"Echo,1,52,64,51,65,55,74,21,1,0" },
    { u8"0072G03", u8"Reverb,1,119,64,20,24,0,3134,102,75,0,49" },
    { u8"0073G",   u8"Dist-4" },
    { u8"0073G00", u8"Overdrive,1,0,64,35,127,61,0,0,9174,21,0,1,1,0" },
    { u8"0073G01", u8"Parametric EQ,1,44,58,39,1404,66,46,17068,24,46,59" },
    { u8"0073G02", u8"Echo,1,119,64,410,65,55,109,21,1,0" },
    { u8"0073G03", u8"Compressor,1,-27,8,-13,32,144,1,1,0,1" },
    { u8"0074G",   u8"Lil Blue" },
    { u8"0074G00", u8"Distorsion,1,0,64,0,8,40,14,1,1785,672,0,1,0,0" },
    { u8"0074G01", u8"Overdrive,1,0,64,35,16,65,14,0,1692,21,0,0,0,0" },
    { u8"0074G02", u8"EQ,1,64,77,74,64,57,59,56,50,47,44,79,61" },
    { u8"0074G03", u8"Reverb,1,116,64,17,6,0,1223,21,97,0,77" },
    { u8"0075G",   u8"Funk-1" },
    { u8"0075G00", u8"Parametric EQ,1,89,64,39,1932,76,82,7691,76,37,64" },
    { u8"0075G01", u8"Chorus,1,62,64,1,0,1,18,0,85,64,119,0,0" },
    { u8"0075G02", u8"Phaser,1,122,64,16,47,0,34,89,9,3,127,1,10" },
    { u8"0075G03", u8"Echo,1,89,64,410,65,55,20,21,1,0" },
    { u8"0076G",   u8"Goog Bye" },
    { u8"0076G00", u8"Parametric EQ,1,93,69,39,825,78,32,11766,37,46,48" },
    { u8"0076G01", u8"Chorus,1,69,64,156,89,0,125,12,2,38,119,0,0" },
    { u8"0076G02", u8"Echo,1,90,64,690,65,55,35,21,1,0" },
    { u8"0076G03", u8"Reverb,1,113,64,22,24,0,4002,27,83,1,64" },
    { u8"0077G",   u8"Naranja One" },
    { u8"0077G00", u8"Parametric EQ,1,72,64,64,600,76,41,8111,64,64,64" },
    { u8"0077G01", u8"Overdrive,1,0,64,35,108,39,0,0,6703,21,0,0,1,0" },
    { u8"0077G02", u8"Echo,1,99,64,566,64,30,59,0,1,0" },
    { u8"0077G03", u8"Chorus,1,72,64,26,0,0,99,43,102,84,119,0,0" },
    { u8"0078G",   u8"Like Pat" },
    { u8"0078G00", u8"Parametric EQ,1,104,78,64,600,80,41,13086,42,36,56" },
    { u8"0078G01", u8"Echo,1,117,64,566,64,30,53,0,1,0" },
    { u8"0078G02", u8"Chorus,1,47,64,39,64,0,23,57,43,75,19,0,0" },
    { u8"0078G03", u8"Reverb,1,110,64,42,10,0,25040,81,78,1,64" },
    { u8"0079G",   u8"The Trooper" },
    { u8"0079G00", u8"Distorsion,1,0,64,0,31,20,0,0,3134,157,0,1,1,0" },
    { u8"0079G01", u8"Reverb,1,91,64,28,24,0,4002,27,83,1,64" },
    { u8"0079G02", u8"Parametric EQ,1,128,51,99,1481,76,64,10032,46,94,87" },
    { u8"0080G",   u8"Hendrix" },
    { u8"0080G00", u8"Distorsion,1,84,64,35,27,34,1,0,824,21,0,0,1,0" },
    { u8"0080G01", u8"Overdrive,1,84,64,35,56,23,0,0,4200,21,0,0,1,0" },
    { u8"0080G02", u8"Reverb,1,111,64,10,26,0,1223,21,77,1,45" },
    { u8"0081G",   u8"Talk-Clean" },
    { u8"0081G00", u8"EQ,1,42,62,42,28,21,35,42,65,53,0,83,94" },
    { u8"0081G01", u8"Echo,1,67,64,347,64,30,59,0,1,0" },
    { u8"0081G02", u8"Chorus,1,64,64,7,0,1,42,97,95,90,127,0,0" },
    { u8"0081G03", u8"WahWah,1,64,64,26,0,0,96,64,0,0,60" },
    { u8"0081G04", u8"Pan,1,64,64,19,0,0,0,10,0,1" },
    { u8"0081G05", u8"Reverb,1,110,64,14,0,0,25040,27,71,0,25" },
    { u8"0081G06", u8"Cabinet,1,1,47" },
    { u8"0082G",   u8"Acoustic" },
    { u8"0082G00", u8"Exciter,1,16,0,0,0,-32,0,-32,-64,0,48,0,26000,20" },
    { u8"0082G01", u8"Sequence,1,20,30,32,20,14,25,30,7,96,120,61,0,0,6,3" },
    { u8"0082G02", u8"Pan,1,0,64,26,0,0,127,32,0,1" },
    { u8"0082G03", u8"Echotron,1,88,64,16,3,0,120,48,64,1,127,-4,64,1,0,0,0" },
    { u8"0083G",   u8"Acoustic Chorus" },
    { u8"0083G00", u8"EQ,1,64,64,64,64,64,64,64,64,64,64,56,64" },
    { u8"0083G01", u8"Exciter,1,64,0,0,0,-32,0,-32,-64,0,48,0,10648,4743" },
    { u8"0083G02", u8"Sequence,1,20,30,32,20,14,25,30,7,72,120,61,0,0,6,3" },
    { u8"0083G03", u8"Reverbtron,1,64,0,1,400,0,0,30,50,6,0,0,64,0,0,20000,0" },
    { u8"0084G",   u8"Chinese Lead" },
    { u8"0084G00", u8"EQ,1,0,0,53,62,67,69,79,81,0,0,65,64" },
    { u8"0084G01", u8"Distorsion,1,0,64,64,127,30,0,0,2316,65,1,1,1,0" },
    { u8"0084G02", u8"Parametric EQ,1,256,46,25,2802,90,53,8111,79,9,16" },
    { u8"0084G03", u8"Flanger,1,64,64,1,0,0,60,102,3,127,64,0,0" },
    { u8"0085G",   u8"Soft Stereo" },
    { u8"0085G00", u8"Parametric EQ,1,72,67,64,1077,36,50,4521,72,50,72" },
    { u8"0085G01", u8"Overdrive,1,84,64,35,22,66,13,1,2982,294,1,1,1,0" },
    { u8"0085G02", u8"Echo,1,34,64,488,105,2,20,0,1,0" },
    { u8"0085G03", u8"Chorus,1,14,64,4,0,0,0,107,40,64,72,0,1" },
    { u8"0085G04", u8"Pan,1,36,64,26,0,0,0,34,0,1" },
    { u8"0085G05", u8"Reverb,1,111,64,30,24,0,25040,118,65,1,47" },
    { u8"0086G",   u8"Acoustic Bright" },
    { u8"0086G00", u8"EQ,1,0,0,109,30,97,18,83,83,51,0,97,25" },
    { u8"0086G01", u8"Distorsion,1,0,64,0,32,68,0,1,3134,157,0,1,1,0" },
    { u8"0086G02", u8"Reverb,1,90,64,51,10,0,25040,81,78,1,64" },
    { u8"0087G",   u8"Rain Echo" },
    { u8"0087G00", u8"EQ,1,0,0,42,95,35,90,88,74,0,0,39,95" },
    { u8"0087G01", u8"Musical Delay,1,48,0,5,1,64,59,23,127,2,69,120,127,127" },
    { u8"0087G02", u8"Flanger,1,64,64,39,0,0,60,23,3,62,0,0,0" },
    { u8"0087G03", u8"Parametric EQ,1,143,79,64,1647,79,64,8111,64,64,44" },
    { u8"0087G04", u8"Cabinet,1,6,64" },
    { u8"0088G",   u8"Swirly" },
    { u8"0088G00", u8"Analog Phaser,1,104,80,5,0,4,64,110,88,5,10,1,97,1" },
    { u8"0088G01", u8"Echo,1,48,64,924,64,23,59,90,1,92" },
    { u8"0088G02", u8"Chorus,1,57,58,9,0,0,69,82,35,56,127,0,0" },
    { u8"0088G03", u8"Parametric EQ,1,29,69,23,1077,42,64,8111,74,64,64" },
    { u8"0089G",   u8"CleanChord" },
    { u8"0089G00", u8"Parametric EQ,1,72,64,64,1077,64,64,8111,64,64,54" },
    { u8"0089G01", u8"Musical Delay,1,43,0,2,7,0,39,9,127,4,32,36,127,127" },
    { u8"0089G02", u8"Chorus,1,64,64,19,0,0,98,56,90,64,19,0,0" },
    { u8"0089G03", u8"Reverb,1,104,64,53,20,0,25040,2417,71,1,64" },
    { u8"0089G04", u8"Cabinet,1,8,64" },
    { u8"0090G",   u8"Shall" },
    { u8"0090G00", u8"Convolotron,1,16,64,1,9,0,0,30,32,0,0,0" },
    { u8"0090G01", u8"DistBand,1,64,64,64,63,24,20,19,26,32,32,24,0,638,1708,0" },
    { u8"0090G02", u8"Parametric EQ,1,412,80,40,3222,88,40,6000,48,40,64" },
    { u8"0090G03", u8"Shelf Boost,1,64,8,6962,1,24" },
    { u8"0090G04", u8"Reverbtron,1,64,0,1,150,0,0,19,48,6,0,0,64,1,0,20000,20" },
    { u8"0090G05", u8"Cabinet,1,4,48" },
    { u8"0091G",   u8"Shall2" },
    { u8"0091G00", u8"Convolotron,1,0,64,1,9,0,0,30,32,0,0,0" },
    { u8"0091G01", u8"StompBox,1,48,3,2,33,16,6" },
    { u8"0091G02", u8"Shelf Boost,1,48,8,13274,0,48" },
    { u8"0091G03", u8"Parametric EQ,1,412,79,25,1651,72,25,8545,69,42,56" },
    { u8"0091G04", u8"Reverbtron,1,64,0,1,100,0,18,0,18,4,0,-2,64,1,0,20000,9" },
    { u8"0091G05", u8"Cabinet,1,2,55" },
    { u8"0092G",   u8"Amp" },
    { u8"0092G00", u8"Sustainer,1,18,127" },
    { u8"0092G01", u8"Shelf Boost,1,55,23,9493,0,52" },
    { u8"0092G02", u8"StompBox,1,49,32,0,32,16,0" },
    { u8"0092G03", u8"Reverbtron,1,88,0,1,141,0,0,67,24,5,0,0,64,1,0,20000,8" },
    { u8"0092G04", u8"Cabinet,1,1,56" },
    { u8"0093G",   u8"Steve Distortion" },
    { u8"0093G00", u8"Compressor,1,-30,2,-6,20,120,1,0,0,1" },
    { u8"0093G01", u8"Valve,1,0,64,64,127,64,0,5078,61,1,0,112,0,30" },
    { u8"0093G02", u8"NoiseGate,1,0,-20,1,2,6703,76,2" },
    { u8"0093G03", u8"StompBox,1,42,0,-17,0,116,6" },
    { u8"0093G04", u8"Reverb,1,100,64,33,0,0,25040,21,106,1,30" },
    { u8"0093G05", u8"Convolotron,1,67,60,1,7,0,0,30,20,4,0,0" },
    { u8"0094G",   u8"Clean Smack Back" },
    { u8"0094G00", u8"Expander,1,-64,22,50,50,8050,76,0" },
    { u8"0094G01", u8"Compressor,1,-34,3,-8,54,130,1,72,0,1" },
    { u8"0094G02", u8"DualFlange,1,-32,0,64,60,10,100,20,16000,0,0,16,90,4,0" },
    { u8"0094G03", u8"Echoverse,1,77,64,105,53,64,76,0,0,0,64" },
    { u8"0094G04", u8"Looper,1,64,1,0,0,1,0,64,0,1,1,64,1,0,0" },
    { u8"0095G",   u8"Soft Dist" },
    { u8"0095G00", u8"DistBand,1,16,64,0,24,48,21,19,21,20,31,25,0,338,800,0" },
    { u8"0095G01", u8"Shelf Boost,1,96,20,14135,0,48" },
    { u8"0095G02", u8"Reverb,1,96,64,24,10,0,6633,5,73,1,74" },
    { u8"0095G03", u8"Parametric EQ,1,207,66,20,2240,68,20,9455,66,1,64" },
    { u8"0096G",   u8"VocoHarm" },
    { u8"0096G00", u8"Vocoder,1,0,64,7,48,92,100,10" },
    { u8"0096G01", u8"Harmonizer,1,83,64,100,8,1201,0,0,0,64,17,0" },
    { u8"0096G02", u8"Reverb,1,80,64,23,24,0,4507,5,83,1,85" },
    { u8"0097G",   u8"Ambience" },
    { u8"0097G00", u8"EQ,1,72,68,64,64,64,64,64,64,68,72,56,40" },
    { u8"0097G01", u8"Echotron,1,112,64,64,4,0,180,64,64,0,127,-24,64,1,0,0,0" },
    { u8"0097G02", u8"Reverbtron,1,88,0,1,220,0,0,38,16,5,0,0,64,1,0,20000,8" },
    { u8"0097G03", u8"Pan,1,32,64,50,0,0,0,32,0,1" },
    { u8"0098G",   u8"Mono Arpeg" },
    { u8"0098G00", u8"Distorsion,1,0,64,0,86,49,27,0,4701,157,0,1,0,0" },
    { u8"0098G01", u8"Ring,1,-42,0,-64,94,74,196,0,55,0,83,26,77,1" },
    { u8"0098G02", u8"Synthfilter,1,76,20,93,0,1,64,110,-40,6,0,0,59,0,500,100,62" },
    { u8"0098G03", u8"Shelf Boost,1,48,29,10979,1,33" },
    { u8"0098G04", u8"Sequence,1,0,40,50,60,70,60,40,0,0,268,64,0,0,5,3" },
    { u8"0098G05", u8"Reverb,1,64,64,33,15,0,5216,5,83,1,74" },
    { u8"0099G",   u8"Padpeg" },
    { u8"0099G00", u8"Synthfilter,1,112,20,90,0,1,64,110,-40,6,0,0,59,0,500,100,62" },
    { u8"0099G01", u8"Shelf Boost,1,48,29,10979,1,25" },
    { u8"0099G02", u8"Sequence,1,0,21,33,59,62,89,106,127,104,540,64,0,0,5,3" },
    { u8"0099G03", u8"Pan,1,64,64,90,0,6,127,32,0,1" },
    { u8"0099G04", u8"Reverb,1,64,64,38,9,0,7342,5,74,0,87" },
    { u8"0100G",   u8"Rock Band" },
    { u8"0100G00", u8"Parametric EQ,1,220,80,16,3484,72,22,7636,72,12,16" },
    { u8"0100G01", u8"Sustainer,1,48,72" },
    { u8"0100G02", u8"Exciter,1,72,0,0,0,0,20,0,-20,0,20,0,4744,120" },
    { u8"0100G03", u8"DistBand,1,0,64,0,32,64,22,19,26,64,32,32,0,400,1200,0" },
    { u8"0100G04", u8"Shelf Boost,1,32,29,10692,0,48" },
    { u8"0100G05", u8"Pan,1,64,64,115,0,5,0,32,0,1" },
    { u8"0100G06", u8"Reverbtron,1,64,11,1,100,0,0,68,22,5,-7,0,64,1,0,20000,8" },
    { u8"0101G",   u8"Rock Band 2" },
    { u8"0101G00", u8"Parametric EQ,1,220,80,16,3484,72,22,7636,72,12,16" },
    { u8"0101G01", u8"Sustainer,1,48,72" },
    { u8"0101G02", u8"Exciter,1,70,0,0,0,0,31,0,-31,0,30,0,4744,120" },
    { u8"0101G03", u8"DistBand,1,0,64,0,24,64,22,19,26,64,32,32,0,400,1200,0" },
    { u8"0101G04", u8"Shelf Boost,1,48,64,16000,0,48" },
    { u8"0101G05", u8"Reverbtron,1,64,11,1,100,0,0,68,22,5,-7,0,64,1,0,20000,8" },
    { u8"0102G",   u8"12 Strings" },
    { u8"0102G00", u8"Shifter,1,94,64,64,165,383,-54,12,0,1,126" },
    { u8"0102G01", u8"Exciter,1,64,0,0,-31,0,29,0,-28,0,22,0,12538,5106" },
    { u8"0103G",   u8"12 Strings Down" },
    { u8"0103G00", u8"Shifter,1,94,64,64,165,383,-54,12,1,1,126" },
    { u8"0103G01", u8"Exciter,1,64,0,0,-31,0,29,0,-28,0,22,0,12538,5106" },
    { u8"0103G02", u8"Echotron,1,107,64,10,3,0,104,32,64,1,96,-8,64,1,0,0,0" },
    { u8"0104G",   u8"12 Strings Crazy" },
    { u8"0104G00", u8"Shifter,1,96,64,64,165,383,-54,12,1,1,126" },
    { u8"0104G01", u8"Exciter,1,64,0,0,-28,0,44,0,-42,0,29,0,6633,1473" },
    { u8"0104G02", u8"Sequence,1,127,0,127,0,127,0,127,0,96,164,64,1,0,5,8" },
    { u8"0104G03", u8"Echotron,1,112,45,34,4,0,164,3,96,0,96,-8,64,1,1,1,1" },
    { u8"0105G",   u8"Stereo Tremolo" },
    { u8"0105G00", u8"Distorsion,1,0,64,0,32,49,27,0,4467,110,0,1,0,0" },
    { u8"0105G01", u8"Chorus,1,64,64,50,0,0,64,50,30,32,64,0,1" },
    { u8"0105G02", u8"Sequence,1,15,48,85,106,125,84,61,33,96,200,95,1,3,4,8" },
    { u8"0105G03", u8"Pan,1,64,64,100,0,5,127,64,1,1" },
    { u8"0105G04", u8"Reverbtron,1,64,0,1,113,0,0,30,20,7,0,0,64,1,0,20000,16" },
    { u8"0105G05", u8"Parametric EQ,1,225,72,37,1258,67,32,6000,64,64,40" },
    { u8"0106G",   u8"Classic WF" },
    { u8"0106G00", u8"StompBox,1,42,64,0,0,110,4" },
    { u8"0106G01", u8"Chorus,1,64,64,60,0,6,127,24,30,32,64,0,1" },
    { u8"0106G02", u8"Sequence,1,4,8,12,8,4,8,12,8,32,120,64,0,0,0,6" },
    { u8"0106G03", u8"Pan,1,32,64,60,127,1,127,96,1,1" },
    { u8"0107G",   u8"Steve Dist 2" },
    { u8"0107G00", u8"NoiseGate,1,-26,-20,1,2,6703,76,2" },
    { u8"0107G01", u8"Coil Crafter,1,45,1,8,3380,10,4466,32,20,1" },
    { u8"0107G02", u8"Compressor,1,-30,2,-6,20,120,1,0,0,1" },
    { u8"0107G03", u8"Valve,1,0,64,64,127,64,0,5078,61,1,0,112,0,30" },
    { u8"0107G04", u8"StompBox,1,45,0,-17,0,116,6" },
    { u8"0107G05", u8"Reverb,1,100,64,33,0,0,25040,21,106,1,30" },
    { u8"0108G",   u8"Auto Wah" },
    { u8"0108G00", u8"Compressor,1,-32,16,-16,10,153,1,0,0,0" },
    { u8"0108G01", u8"Shifter,1,64,0,64,0,0,-20,1,1,1,20" },
    { u8"0108G02", u8"Valve,1,0,64,0,24,48,0,3200,110,0,1,0,0,80" },
    { u8"0108G03", u8"Sequence,1,11,55,15,77,13,64,11,36,96,60,110,0,4,1,2" },
    { u8"0108G04", u8"Pan,1,0,64,26,0,0,0,96,0,1" },
    { u8"0108G05", u8"VaryBand,1,32,120,0,64,240,6,127,400,1400,2000,1" },
    { u8"0109G",   u8"Auto Wah 2" },
    { u8"0109G00", u8"Compressor,1,-33,4,-16,10,153,1,0,0,0" },
    { u8"0109G01", u8"Shifter,1,64,0,64,0,0,-20,1,1,1,20" },
    { u8"0109G02", u8"Sequence,1,0,0,96,96,0,0,96,96,16,80,104,0,4,0,2" },
    { u8"0109G03", u8"Pan,1,0,64,240,0,0,0,64,0,1" },
    { u8"0110G",   u8"Stereo Vibrato" },
    { u8"0110G00", u8"Convolotron,1,16,64,0,20,0,0,30,24,0,0,0" },
    { u8"0110G01", u8"Sequence,1,64,30,45,20,60,25,42,15,64,120,64,0,1,6,3" },
    { u8"0110G02", u8"VaryBand,1,0,200,5,127,200,6,0,400,1000,2000,1" },
    { u8"0110G03", u8"Exciter,1,64,0,0,0,0,3,0,-64,0,64,0,20000,20" },
    { u8"0110G04", u8"Pan,1,32,64,200,0,6,126,64,1,1" },
    { u8"0111G",   u8"Dist Dead" },
    { u8"0111G00", u8"Compressor,1,-36,10,-8,10,213,1,26,0,1" },
    { u8"0111G01", u8"StompBox,1,48,0,6,6,24,5" },
    { u8"0111G02", u8"Shuffle,1,32,-5,3,-64,-57,252,1251,3000,6000,-22,1" },
    { u8"0111G03", u8"VaryBand,1,32,234,0,127,234,6,0,400,2600,4000,1" },
    { u8"0111G04", u8"Pan,1,0,64,234,0,0,127,96,0,1" },
    { u8"0111G05", u8"Echotron,1,96,35,34,4,0,58,3,128,0,127,-4,64,1,1,0,1" },
    { u8"0112G",   u8"Rhythmic Vibes" },
    { u8"0112G00", u8"Expander,1,-55,37,24,46,8050,80,1" },
    { u8"0112G01", u8"StompBox,1,44,64,-1,23,9,4" },
    { u8"0112G02", u8"VaryBand,1,0,140,0,32,80,4,0,465,1764,3309,1" },
    { u8"0113G",   u8"Clean Tremolo" },
    { u8"0113G00", u8"Convolotron,1,32,64,0,50,0,0,30,31,0,0,0" },
    { u8"0113G01", u8"Pan,1,0,64,340,0,1,64,0,1,0" },
    { u8"0114G",   u8"Double Tremolo" },
    { u8"0114G00", u8"Convolotron,1,32,64,0,50,0,0,18,16,0,0,0" },
    { u8"0114G01", u8"VaryBand,1,16,70,1,64,140,1,64,514,2324,3946,8" },
    { u8"0114G02", u8"Pan,1,32,64,350,0,5,64,96,1,1" },
    { u8"0115G",   u8"Stereo" },
    { u8"0115G00", u8"EQ,1,64,64,64,64,64,64,64,64,64,64,56,64" },
    { u8"0115G01", u8"Shifter,1,64,64,64,0,0,-20,1,0,1,22" },
    { u8"0115G02", u8"Exciter,1,64,0,0,0,0,0,0,-20,0,20,0,15608,100" },
    { u8"0115G03", u8"VaryBand,1,16,160,1,127,80,5,0,500,2500,5000,1" },
    { u8"0115G04", u8"Pan,1,32,64,26,0,0,0,96,0,1" },
    { u8"0115G05", u8"Reverbtron,1,88,0,1,400,0,30,70,22,5,0,0,64,1,0,20000,10" },
    { u8"0116G",   u8"Circle Tremolo" },
    { u8"0116G00", u8"Parametric EQ,1,24,64,64,75,64,64,113,64,64,32" },
    { u8"0116G01", u8"Sequence,1,8,40,66,92,105,82,54,15,32,140,64,1,4,4,8" },
    { u8"0116G02", u8"VaryBand,1,32,70,5,0,280,1,127,400,1000,4162,0" },
    { u8"0116G03", u8"Pan,1,96,64,35,0,1,127,96,1,1" },
    { u8"0117G",   u8"Rhythm" },
    { u8"0117G00", u8"Shifter,1,64,64,64,165,165,-20,0,1,0,20" },
    { u8"0117G01", u8"Compressor,1,-30,2,-12,20,120,1,0,0,0" },
    { u8"0117G02", u8"Sequence,1,28,59,94,127,120,80,50,24,64,60,59,0,4,0,6" },
    { u8"0117G03", u8"MuTroMojo,1,64,1,60,0,6,64,32,34,96,96,16,-16,16,3,4000,400,1,0" },
    { u8"0117G04", u8"Pan,1,48,64,60,0,4,96,96,1,1" },
    { u8"0118G",   u8"Bright Size" },
    { u8"0118G00", u8"Compressor,1,-20,8,-16,20,120,1,0,0,0" },
    { u8"0118G01", u8"Sequence,1,0,21,33,59,62,89,106,127,106,540,64,0,0,5,6" },
    { u8"0118G02", u8"Pan,1,32,64,90,0,6,127,32,0,1" },
    { u8"0118G03", u8"Reverb,1,64,64,38,15,0,7342,747,90,0,80" },
    { u8"0119G",   u8"Metal" },
    { u8"0119G00", u8"Sustainer,1,48,100" },
    { u8"0119G01", u8"StompBox,1,48,2,2,2,8,6" },
    { u8"0119G02", u8"Sequence,1,9,35,60,67,67,62,46,16,80,100,88,0,4,0,8" },
    { u8"0119G03", u8"Shelf Boost,1,48,0,9600,0,48" },
    { u8"0119G04", u8"Reverb,1,64,64,18,8,0,4800,200,64,0,72" },
    { u8"0120G",   u8"Floyd Trip" },
    { u8"0120G00", u8"Compressor,1,-13,2,-10,20,120,1,0,0,0" },
    { u8"0120G01", u8"Sequence,1,0,21,33,59,62,89,106,127,67,480,64,0,3,5,6" },
    { u8"0120G02", u8"Synthfilter,1,0,5,60,0,0,0,53,-12,4,0,1,21,17,313,158,127" },
    { u8"0120G03", u8"Reverb,1,90,64,35,5,7,7342,170,64,0,83" },
    { u8"0121G",   u8"Dreamland" },
    { u8"0121G00", u8"EQ,1,64,64,65,59,60,65,67,70,67,64,64,37" },
    { u8"0121G01", u8"Sustainer,1,35,67" },
    { u8"0121G02", u8"Overdrive,1,0,64,64,2,74,28,0,9467,120,0,1,0,0" },
    { u8"0121G03", u8"Analog Phaser,1,67,0,120,50,1,32,23,34,10,27,0,99,1" },
    { u8"0121G04", u8"Echotron,1,93,64,127,4,0,99,3,41,0,96,-13,64,1,0,0,0" },
    { u8"0122G",   u8"Old Tavern" },
    { u8"0122G00", u8"Sustainer,1,8,91" },
    { u8"0122G01", u8"Analog Phaser,1,64,60,180,50,0,48,79,100,1,0,1,53,1" },
    { u8"0122G02", u8"Reverbtron,1,64,25,1,128,0,0,0,72,0,2,-29,68,1,0,20000,88" },
    { u8"0123G",   u8"Mild Clean" },
    { u8"0123G00", u8"Coil Crafter,1,32,5,3,2700,21,2900,18,179,1" },
    { u8"0123G01", u8"Sustainer,1,74,54" },
    { u8"0123G02", u8"Echotron,1,40,102,13,8,0,160,0,64,3,32,0,64,1,0,0,0" },
    { u8"0123G03", u8"Analog Phaser,1,83,15,10,50,0,0,79,33,2,0,1,53,1" },
    { u8"0123G04", u8"Reverbtron,1,81,65,0,316,0,0,0,32,3,0,-10,68,1,0,965,46" },
    { u8"0124G",   u8"Arppeg Sprinkler" },
    { u8"0124G00", u8"Coil Crafter,1,32,5,3,2700,21,2900,18,179,1" },
    { u8"0124G01", u8"Sustainer,1,74,54" },
    { u8"0124G02", u8"Arpie,1,67,64,35,64,30,59,0,127,0,0,0" },
    { u8"0124G03", u8"Analog Phaser,1,83,15,10,50,0,0,79,33,2,0,1,53,1" },
    { u8"0124G04", u8"Echotron,1,40,102,13,8,0,160,0,64,3,32,0,64,1,0,0,0" },
    { u8"0124G05", u8"Reverbtron,1,81,65,0,316,0,0,0,51,3,0,-10,68,1,0,965,46" },
    { u8"0125G",   u8"Backward Tumble" },
    { u8"0125G00", u8"Coil Crafter,1,32,5,3,2700,21,2900,18,179,0" },
    { u8"0125G01", u8"Sustainer,1,95,54" },
    { u8"0125G02", u8"DualFlange,1,-15,0,0,42,12,50,-10,1500,0,0,120,0,0,20" },
    { u8"0125G03", u8"Echoverse,1,60,64,70,64,64,16,39,127,0,0" },
    { u8"0125G04", u8"VaryBand,1,25,50,0,32,30,4,0,305,1764,3309,2" },
    { u8"0125G05", u8"Echotron,1,81,53,34,4,0,70,3,41,0,96,-13,64,1,0,1,1" },
    { u8"0125G06", u8"Reverbtron,1,81,65,0,316,0,0,0,51,3,0,-10,68,1,0,965,46" },
    { u8"0126G",   u8"Dig For Shift" },
    { u8"0126G00", u8"Sustainer,1,95,26" },
    { u8"0126G01", u8"Shifter,1,64,64,64,20,20,-5,5,0,0,0" },
    { u8"0126G02", u8"Echotron,1,60,49,46,8,0,59,0,64,3,0,0,64,1,0,0,0" },
    { u8"0126G03", u8"Reverbtron,1,64,0,1,400,0,0,99,21,0,0,0,64,0,0,20000,0" },
    { u8"0127G",   u8"Diode Overdose" },
    { u8"0127G00", u8"Distorsion,1,16,64,64,23,43,29,0,4744,720,0,1,0,0" },
    { u8"0127G01", u8"Overdrive,1,25,64,64,0,58,29,0,722,20,0,0,0,0" },
    { u8"0127G02", u8"Echoverse,1,109,64,107,18,30,67,25,0,0,81" },
    { u8"0128G",   u8"Warm Chorus" },
    { u8"0128G00", u8"Coil Crafter,1,44,5,2,3153,34,3464,51,179,0" },
    { u8"0128G01", u8"Sustainer,1,104,38" },
    { u8"0128G02", u8"Analog Phaser,1,83,15,10,50,0,0,79,33,2,0,1,53,1" },
    { u8"0128G03", u8"DualFlange,1,-15,0,0,46,14,50,-10,1500,0,0,120,0,0,20" },
    { u8"0128G04", u8"Shuffle,1,30,36,-13,-13,36,60,858,2436,6000,-24,1" },
    { u8"0128G05", u8"Reverbtron,1,86,37,0,881,0,0,0,17,3,0,-25,68,1,0,965,127" },
    { u8"0129G",   u8"SeqDist" },
    { u8"0129G00", u8"Sustainer,1,30,38" },
    { u8"0129G01", u8"StompBox,1,32,64,0,0,110,4" },
    { u8"0129G02", u8"Sequence,1,0,0,62,0,51,21,5,64,0,219,95,1,3,4,2" },
    { u8"0129G03", u8"Vocoder,1,0,64,31,73,97,95,16" },
    { u8"0129G04", u8"Reverbtron,1,79,0,1,128,0,0,99,30,0,0,0,64,0,0,20000,58" },
    { u8"0130G",   u8"Music" },
    { u8"0130G00", u8"CompBand,1,64,8,4,4,4,-21,16,16,-21,100,1063,7189,64" },
    { u8"0130G01", u8"Compressor,1,-60,2,0,10,500,1,0,1,1" },
    { u8"0131G",   u8"C Melody" },
    { u8"0131G00", u8"Ring,1,50,0,-64,127,38,698,1,95,0,73,0,85,1" },
    { u8"0131G01", u8"Harmonizer,1,64,64,64,17,64,1,0,0,64,64,0" },
    { u8"0131G02", u8"CompBand,1,64,8,2,2,4,6,10,9,-12,691,1063,7189,64" },
    { u8"0131G03", u8"Reverb,1,96,64,18,16,0,6218,220,78,1,83" },
    { u8"0131G04", u8"Pan,1,32,64,50,0,0,0,96,0,1" },
    { u8"0132G",   u8"Surf's Up" },
    { u8"0132G00", u8"CompBand,1,26,2,4,3,4,-32,24,-28,-14,383,1000,5055,55" },
    { u8"0132G01", u8"OpticalTrem,1,60,217,10,0,64,64" },
    { u8"0132G02", u8"Overdrive,1,0,64,64,2,75,28,0,4255,21,1,0,0,0" },
    { u8"0132G03", u8"Chorus,1,32,64,120,0,8,90,31,85,64,0,0,1" },
    { u8"0132G04", u8"Convolotron,1,0,64,0,90,0,0,0,85,5,0,0" },
    { u8"0132G05", u8"Reverbtron,1,100,0,1,155,0,30,0,0,9,0,0,64,1,0,2854,60" },
    { u8"0133G",   u8"Live Punk Show" },
    { u8"0133G00", u8"Expander,1,-70,33,18,46,8050,80,1" },
    { u8"0133G01", u8"Distorsion,1,49,64,0,30,44,29,0,3134,680,0,1,0,0" },
    { u8"0133G02", u8"Overdrive,1,0,64,64,59,37,28,0,4255,150,1,1,0,0" },
    { u8"0133G03", u8"Convolotron,1,0,75,0,93,0,0,55,77,2,0,0" },
    { u8"0133G04", u8"Reverbtron,1,63,0,1,155,0,30,0,0,9,0,0,64,1,0,2854,60" },
    { u8"0134G",   u8"Charred Violets" },
    { u8"0134G00", u8"Expander,1,-71,31,50,50,8050,76,1" },
    { u8"0134G01", u8"Vibe,1,35,240,10,0,0,64,64,64,3,64" },
    { u8"0134G02", u8"StompBox,1,17,-47,5,-26,127,2" },
    { u8"0134G03", u8"Echotron,1,62,78,37,5,0,60,52,64,4,96,-2,64,1,0,0,1" },
    { u8"0135G",   u8"Stereo Ecstasy" },
    { u8"0135G00", u8"CompBand,1,26,2,4,6,4,-34,24,-28,-11,431,1000,5055,55" },
    { u8"0135G01", u8"Overdrive,1,0,64,64,2,75,28,0,4255,21,1,0,0,0" },
    { u8"0135G02", u8"Echotron,1,64,36,93,6,0,45,0,64,8,32,0,64,0,0,1,1" },
    { u8"0135G03", u8"DualFlange,1,-15,0,0,42,12,50,-10,1500,0,0,105,0,8,20" },
    { u8"0135G04", u8"VaryBand,1,0,80,0,64,40,0,0,120,1000,2300,1" },
    { u8"0135G05", u8"Reverbtron,1,100,0,1,155,0,30,0,0,9,0,0,64,1,0,2854,60" },
    { u8"0136G",   u8"Insanity Mojo" },
    { u8"0136G00", u8"Expander,1,-70,33,18,46,8050,80,1" },
    { u8"0136G01", u8"MuTroMojo,1,0,22,138,0,0,64,0,36,0,20,32,0,5,1,3735,60,0,1" },
    { u8"0136G02", u8"Overdrive,1,0,64,64,59,37,28,0,4255,150,1,1,0,0" },
    { u8"0137G",   u8"Random Eyes" },
    { u8"0137G00", u8"Expander,1,-70,33,18,46,8050,80,1" },
    { u8"0137G01", u8"Distorsion,1,49,64,0,30,44,29,0,3134,680,0,1,0,0" },
    { u8"0137G02", u8"Overdrive,1,0,64,64,67,37,28,0,2943,20,1,0,0,0" },
    { u8"0137G03", u8"MuTroMojo,1,0,12,145,0,9,64,59,5,15,40,32,0,-2,3,3125,60,0,1" },
    { u8"0137G04", u8"Reverbtron,1,64,0,1,400,0,0,53,25,0,0,0,64,0,0,20000,0" },
    { u8"0137G05", u8"Parametric EQ,1,153,75,47,1689,57,25,8927,53,40,64" },
    { u8"0138G",   u8"Goin through a phase" },
    { u8"0138G00", u8"Expander,1,-70,33,18,46,8050,80,1" },
    { u8"0138G01", u8"Distorsion,1,49,64,0,30,44,29,0,3134,680,0,1,0,0" },
    { u8"0138G02", u8"Overdrive,1,0,64,64,67,37,28,0,2943,20,1,0,0,0" },
    { u8"0138G03", u8"Analog Phaser,1,64,20,145,50,9,64,92,40,6,0,1,58,1" },
    { u8"0138G04", u8"Reverbtron,1,64,0,1,400,0,0,53,25,0,0,0,64,0,0,20000,0" },
    { u8"0138G05", u8"EQ,1,64,64,65,59,60,65,67,70,67,64,64,37" },
    { u8"0138G06", u8"Parametric EQ,1,153,75,47,1689,61,25,12829,58,40,64" },
    { u8"0139G",   u8"Suction II" },
    { u8"0139G00", u8"Expander,1,-64,22,50,50,1200,70,0" },
    { u8"0139G01", u8"Sustainer,1,75,73" },
    { u8"0139G02", u8"Overdrive,1,0,64,35,4,51,28,0,2382,21,1,0,0,0" },
    { u8"0139G03", u8"Echotron,1,32,46,37,8,0,92,0,64,3,75,0,64,1,1,0,1" },
    { u8"0139G04", u8"Reverbtron,1,104,0,1,400,0,0,60,15,2,0,0,64,0,0,20000,20" },
    { u8"0139G05", u8"Reverb,1,48,64,13,18,0,4002,27,83,0,64" },
    { u8"0140G",   u8"Multichorus" },
    { u8"0140G00", u8"Sustainer,1,35,113" },
    { u8"0140G01", u8"Echotron,1,79,64,42,7,0,60,0,61,7,32,0,64,1,1,0,1" },
    { u8"0141G",   u8"Clean Twirl" },
    { u8"0141G00", u8"Expander,1,-64,22,50,50,8050,76,0" },
    { u8"0141G01", u8"Sustainer,1,39,91" },
    { u8"0141G02", u8"Analog Phaser,1,64,20,32,50,1,64,93,40,6,48,0,58,1" },
    { u8"0141G03", u8"Echoverse,1,111,64,90,64,64,18,23,0,1,82" },
    { u8"0141G04", u8"Reverb,1,116,64,62,24,0,3134,102,75,1,49" },
    { u8"0142G",   u8"TriadEchoes" },
    { u8"0142G00", u8"Expander,1,-64,22,50,50,8050,76,0" },
    { u8"0142G01", u8"StompBox,1,42,-20,0,0,70,2" },
    { u8"0142G02", u8"Echotron,1,98,72,37,3,0,73,3,47,0,96,-6,0,1,0,0,1" },
    { u8"0142G03", u8"Echo,1,51,127,500,64,64,35,10,0,0" },
    { u8"0143G",   u8"GibSoonMur" },
    { u8"0143G00", u8"Expander,1,-55,30,50,50,1441,157,50" },
    { u8"0143G01", u8"Compressor,1,-26,4,-8,20,120,1,10,1,1" },
    { u8"0143G02", u8"Coil Crafter,1,32,1,6,4400,42,3300,17,20,0" },
    { u8"0143G03", u8"Echotron,1,20,65,34,2,0,96,3,41,0,96,-13,64,1,1,1,0" },
    { u8"0143G04", u8"Parametric EQ,1,72,71,38,1077,64,64,10580,69,38,64" },
    { u8"0144G",   u8"GibSoonMurDouble" },
    { u8"0144G00", u8"Expander,1,-55,30,50,50,1441,157,50" },
    { u8"0144G01", u8"Compressor,1,-26,4,-8,20,120,1,10,0,0" },
    { u8"0144G02", u8"Coil Crafter,1,10,1,6,4400,42,3300,17,20,1" },
    { u8"0144G03", u8"Echotron,1,20,65,34,2,0,63,3,41,0,96,-13,64,1,1,1,0" },
    { u8"0144G04", u8"Parametric EQ,1,72,71,38,1077,64,64,10580,69,38,64" },
    { u8"0145G",   u8"GibSoonMurder" },
    { u8"0145G00", u8"Expander,1,-55,30,50,50,1441,157,50" },
    { u8"0145G01", u8"Compressor,1,-26,4,-8,20,120,1,10,1,1" },
    { u8"0145G02", u8"Coil Crafter,1,10,1,6,4400,42,3300,17,20,0" },
    { u8"0145G03", u8"Echotron,1,20,65,34,4,0,96,3,41,0,96,-13,64,1,1,4,0" },
    { u8"0145G04", u8"Parametric EQ,1,72,71,38,1077,64,64,10580,69,38,64" },
    { u8"0146G",   u8"Igniter" },
    { u8"0146G00", u8"Expander,1,-51,50,59,46,8050,76,1" },
    { u8"0146G01", u8"StompBox,1,38,6,6,6,105,1" },
    { u8"0146G02", u8"Convolotron,1,0,64,0,76,0,0,0,62,0,0,0" },
    { u8"0146G03", u8"Reverbtron,1,64,0,0,327,0,0,0,0,7,0,0,64,0,0,1585,12" },
    { u8"0147G",   u8"Ignited" },
    { u8"0147G00", u8"Expander,1,-65,15,10,300,8000,20,30" },
    { u8"0147G01", u8"StompBox,1,60,0,-12,-9,85,6" },
    { u8"0147G02", u8"Convolotron,1,0,64,0,76,0,0,0,74,3,0,0" },
    { u8"0147G03", u8"Overdrive,1,0,64,64,0,79,28,1,10884,170,1,1,0,0" },
    { u8"0147G04", u8"Reverbtron,1,88,0,1,400,0,0,88,14,5,0,0,64,0,0,20000,0" },
    { u8"0148G",   u8"BlowMoeJo" },
    { u8"0148G00", u8"EQ,1,71,68,64,64,64,64,64,64,66,69,64,40" },
    { u8"0148G01", u8"Compressor,1,-30,2,-13,20,120,1,0,0,0" },
    { u8"0148G02", u8"OpticalTrem,1,127,480,0,3,32,64" },
    { u8"0148G03", u8"Overdrive,1,84,64,35,56,40,0,0,7600,21,0,1,0,0" },
    { u8"0148G04", u8"MuTroMojo,1,16,10,60,0,1,64,0,29,0,0,-16,40,-3,1,2000,450,1,1" },
    { u8"0149G",   u8"JoeMoeJo" },
    { u8"0149G00", u8"EQ,1,71,68,64,64,64,64,64,64,66,69,64,40" },
    { u8"0149G01", u8"Compressor,1,-30,2,-12,20,120,1,0,0,0" },
    { u8"0149G02", u8"MuTroMojo,1,0,15,138,0,0,64,0,39,0,30,32,0,5,1,2000,60,0,1" },
    { u8"0150G",   u8"Gary Moded" },
    { u8"0150G00", u8"EQ,1,18,40,53,58,68,66,61,45,26,0,89,68" },
    { u8"0150G01", u8"Parametric EQ,1,55,28,72,704,65,104,5896,52,64,92" },
    { u8"0150G02", u8"Sustainer,1,21,127" },
    { u8"0150G03", u8"Overdrive,1,84,64,35,56,40,0,0,7600,21,0,1,0,0" },
    { u8"0150G04", u8"Echo,1,0,64,272,64,64,35,0,1,0" },
    { u8"0151G",   u8"Gary Loaded" },
    { u8"0151G00", u8"Expander,1,-45,50,59,250,8050,76,1" },
    { u8"0151G01", u8"Sustainer,1,16,127" },
    { u8"0151G02", u8"Parametric EQ,1,55,28,72,704,65,104,5896,52,64,90" },
    { u8"0151G03", u8"Overdrive,1,84,64,35,56,40,0,0,7600,21,0,1,0,0" },
    { u8"0151G04", u8"Echo,1,37,64,565,75,30,59,1,0,0" },
    { u8"0152G",   u8"Oxygen" },
    { u8"0152G00", u8"Vibe,1,75,330,10,0,50,64,0,64,17,64" },
    { u8"0152G01", u8"Reverb,1,48,64,13,18,0,4002,27,83,0,64" },
    { u8"0153G",   u8"Oxygenated" },
    { u8"0153G00", u8"EQ,1,18,40,53,58,68,66,61,45,26,0,89,68" },
    { u8"0153G01", u8"Overdrive,1,0,64,35,110,23,13,0,2982,21,0,0,0,0" },
    { u8"0153G02", u8"Sustainer,1,16,127" },
    { u8"0153G03", u8"Parametric EQ,1,55,28,72,704,65,104,5896,52,64,92" },
    { u8"0153G04", u8"Echo,1,0,64,488,64,64,35,0,1,0" },
    { u8"0154G",   u8"StartButton" },
    { u8"0154G00", u8"Expander,1,-51,50,59,46,8050,76,1" },
    { u8"0154G01", u8"StompBox,1,72,6,6,6,105,6" },
    { u8"0154G02", u8"Convolotron,1,0,64,0,76,0,0,0,53,0,0,0" },
    { u8"0154G03", u8"Overdrive,1,0,64,64,0,79,28,1,10884,170,1,1,0,0" },
    { u8"0155G",   u8"MetalAmploaded" },
    { u8"0155G00", u8"Expander,1,-51,50,59,46,8050,76,1" },
    { u8"0155G01", u8"StompBox,1,48,0,-32,-36,53,6" },
    { u8"0155G02", u8"Convolotron,1,0,64,0,76,0,0,0,70,0,0,0" },
    { u8"0155G03", u8"Reverbtron,1,64,0,0,327,0,0,0,0,7,0,0,64,0,0,1585,12" },
    { u8"0156G",   u8"MetalAmpReloaded" },
    { u8"0156G00", u8"Expander,1,-55,30,50,50,1441,157,50" },
    { u8"0156G01", u8"StompBox,1,74,0,-12,0,127,6" },
    { u8"0156G02", u8"Reverbtron,1,64,0,1,400,0,0,30,34,6,0,0,64,0,0,20000,0" },
    { u8"0156G03", u8"Convolotron,1,67,64,0,100,0,0,30,20,8,0,0" },
    { u8"0157G",   u8"MetalAmpOverloaded" },
    { u8"0157G00", u8"Expander,1,-65,31,10,50,1441,157,50" },
    { u8"0157G01", u8"Overdrive,1,18,64,35,25,48,29,1,2854,336,0,1,0,0" },
    { u8"0157G02", u8"StompBox,1,90,-8,-17,0,127,6" },
    { u8"0157G03", u8"Convolotron,1,0,64,0,100,0,0,25,60,8,0,0" },
    { u8"0157G04", u8"Reverbtron,1,64,0,1,400,0,0,30,34,6,0,0,64,0,0,20000,0" },
    { u8"0158G",   u8"Melon Collie" },
    { u8"0158G00", u8"Sustainer,1,88,67" },
    { u8"0158G01", u8"Vibe,1,67,260,10,0,48,70,20,35,0,48" },
    { u8"0158G02", u8"OpticalTrem,1,51,130,10,0,80,64" },
    { u8"0158G03", u8"Reverbtron,1,64,25,1,128,0,0,0,72,0,2,-29,68,1,0,20000,88" },
    { u8"0159G",   u8"Melon Collie II" },
    { u8"0159G00", u8"Sustainer,1,88,67" },
    { u8"0159G01", u8"Vibe,1,67,260,10,0,48,70,20,35,0,48" },
    { u8"0159G02", u8"OpticalTrem,1,35,265,10,0,80,64" },
    { u8"0159G03", u8"Reverbtron,1,64,25,1,24,0,0,0,72,0,2,-29,68,1,0,20000,88" },
    { u8"0160G",   u8"Sounds on Sound" },
    { u8"0160G00", u8"Vibe,1,51,12,10,9,63,64,0,14,0,30" },
    { u8"0160G01", u8"Reverbtron,1,63,0,1,20,0,0,0,53,8,27,0,68,1,0,492,6" },
    { u8"0161G",   u8"Voy Pa Alla" },
    { u8"0161G00", u8"Sequence,1,11,55,15,70,12,60,11,36,79,70,92,0,4,1,8" },
    { u8"0161G01", u8"OpticalTrem,1,80,70,0,6,0,64" },
    { u8"0161G02", u8"WahWah,1,64,64,70,0,0,127,64,0,0,60" },
    { u8"0161G03", u8"Pan,1,32,64,50,0,0,0,97,0,1" },
    { u8"0161G04", u8"Reverb,1,80,64,63,24,0,85,5,83,1,64" },
    { u8"0162G",   u8"JimiVibe" },
    { u8"0162G00", u8"Vibe,1,46,120,10,0,48,64,0,64,3,64" },
    { u8"0162G01", u8"Valve,1,9,64,64,120,37,0,5688,483,1,1,106,0,100" },
    { u8"0162G02", u8"StompBox,1,48,1,3,17,0,7" },
    { u8"0162G03", u8"Reverbtron,1,64,16,1,208,0,0,0,42,3,-17,0,64,0,0,20000,0" },
    { u8"0163G",   u8"JimiVibe2" },
    { u8"0163G00", u8"Vibe,1,60,197,10,0,48,64,64,64,0,64" },
    { u8"0163G01", u8"Valve,1,9,64,64,120,37,0,5688,273,1,1,72,0,78" },
    { u8"0163G02", u8"StompBox,1,48,1,3,-34,0,7" },
    { u8"0163G03", u8"Reverbtron,1,64,16,1,208,0,0,0,42,3,-17,0,64,0,0,20000,0" },
    { u8"0164G",   u8"Sounds and-a-log" },
    { u8"0164G00", u8"Sustainer,1,60,22" },
    { u8"0164G01", u8"Vibe,1,110,75,10,0,32,64,64,14,0,30" },
    { u8"0164G02", u8"Valve,1,60,64,64,111,62,0,5841,747,1,1,28,0,91" },
    { u8"0164G03", u8"Echotron,1,64,65,34,3,0,76,40,65,0,96,-38,64,1,1,8,0" },
    { u8"0164G04", u8"Reverbtron,1,64,25,1,128,0,0,0,72,0,2,-29,68,1,0,20000,88" },
    { u8"0165G",   u8"Envelope YaYa" },
    { u8"0165G00", u8"EQ,1,64,62,64,64,68,70,72,74,66,69,64,40" },
    { u8"0165G01", u8"MuTroMojo,1,0,31,138,0,0,64,0,24,0,28,29,0,0,2,5455,60,0,1" },
    { u8"0165G02", u8"Ring,1,-64,0,0,65,100,1820,1,7,36,0,35,81,0" },
    { u8"0165G03", u8"Parametric EQ,1,145,74,30,1077,64,64,12182,51,25,64" },
    { u8"0166G",   u8"AutoYa" },
    { u8"0166G00", u8"EQ,1,64,62,64,64,68,70,65,69,66,69,64,64" },
    { u8"0166G01", u8"Compressor,1,-30,2,-12,20,120,1,0,0,0" },
    { u8"0166G02", u8"MuTroMojo,1,0,31,120,0,0,64,28,-5,46,42,29,0,0,2,5455,60,0,1" },
    { u8"0166G03", u8"Ring,1,-50,0,0,76,100,2320,1,7,36,0,35,81,0" },
    { u8"0166G04", u8"Reverb,1,97,64,63,24,0,85,5,83,1,64" },
    { u8"0166G05", u8"Parametric EQ,1,145,70,30,1077,64,64,12182,58,25,64" },
    { u8"0167G",   u8"AutoYiYa" },
    { u8"0167G00", u8"EQ,1,64,62,64,64,68,70,72,74,66,69,64,40" },
    { u8"0167G01", u8"MuTroMojo,1,0,31,45,0,4,64,23,-4,62,28,29,0,0,2,5455,60,0,1" },
    { u8"0167G02", u8"Ring,1,-64,0,0,65,100,1820,1,7,36,0,35,81,0" },
    { u8"0167G03", u8"Parametric EQ,1,145,74,30,1077,64,64,12182,51,25,64" },
    { u8"0168G",   u8"RandYa" },
    { u8"0168G00", u8"EQ,1,64,62,64,64,68,70,65,69,66,69,64,64" },
    { u8"0168G01", u8"Compressor,1,-30,2,-12,20,120,1,0,0,0" },
    { u8"0168G02", u8"MuTroMojo,1,0,31,120,0,9,64,31,-5,46,42,29,0,0,2,5455,60,0,1" },
    { u8"0168G03", u8"Ring,1,-52,0,0,76,100,2320,1,7,36,0,35,81,0" },
    { u8"0168G04", u8"Reverb,1,97,64,63,24,0,85,5,83,1,64" },
    { u8"0168G05", u8"Parametric EQ,1,145,70,30,1077,64,64,12182,58,25,64" },
    { u8"0169G",   u8"Talk to me" },
    { u8"0169G00", u8"EQ,1,64,62,64,64,68,70,65,69,66,69,64,64" },
    { u8"0169G01", u8"Compressor,1,-31,6,-10,20,120,1,22,0,1" },
    { u8"0169G02", u8"MuTroMojo,1,0,31,208,0,8,64,37,-6,28,42,29,0,0,2,5455,60,0,1" },
    { u8"0169G03", u8"Ring,1,-45,0,0,67,100,2550,1,7,36,0,35,81,0" },
    { u8"0169G04", u8"Reverb,1,97,64,63,24,0,85,5,83,1,64" },
    { u8"0169G05", u8"Parametric EQ,1,145,70,30,1077,64,64,12182,58,25,64" },
    { u8"0170G",   u8"Helicopter" },
    { u8"0170G00", u8"Expander,1,-80,37,24,46,8050,76,1" },
    { u8"0170G01", u8"StompBox,1,55,27,10,1,18,0" },
    { u8"0170G02", u8"OpticalTrem,1,127,556,0,2,64,64" },
    { u8"0170G03", u8"Valve,1,0,64,64,30,67,0,3327,20,1,1,46,1,67" },
    { u8"0170G04", u8"VaryBand,1,32,3,0,53,6,0,72,269,1509,3309,2" },
    { u8"0170G05", u8"DualFlange,1,0,0,0,107,2291,98,3,12371,0,0,3,64,1,10" },
    { u8"0171G",   u8"Octaflange" },
    { u8"0171G00", u8"EQ,1,78,71,64,64,64,64,64,64,71,80,64,42" },
    { u8"0171G01", u8"Flanger,1,64,64,14,0,1,62,12,19,97,0,0,0" },
    { u8"0171G02", u8"Harmonizer,1,64,64,64,0,6000,0,0,0,64,64,0" },
    { u8"0171G03", u8"Compressor,1,-30,4,-6,20,120,1,100,1,1" },
    { u8"0172G",   u8"Octaflange - Dirty" },
    { u8"0172G00", u8"Expander,1,-65,35,10,117,6633,20,1" },
    { u8"0172G01", u8"Flanger,1,64,64,14,0,1,62,12,19,97,0,0,0" },
    { u8"0172G02", u8"Harmonizer,1,64,64,64,0,6000,0,0,0,64,64,0" },
    { u8"0172G03", u8"Valve,1,9,64,64,80,64,1,5688,220,1,1,58,1,30" },
    { u8"0172G04", u8"Compressor,1,-30,4,-6,20,120,1,100,1,1" },
    { u8"0172G05", u8"EQ,1,69,71,71,64,65,67,72,65,58,40,64,55" },
    { u8"0173G",   u8"OctaVibe - Dirty Echo" },
    { u8"0173G00", u8"Expander,1,-65,35,10,117,6633,20,1" },
    { u8"0173G01", u8"Exciter,1,64,64,-62,0,0,64,0,-45,0,64,0,20000,5106" },
    { u8"0173G02", u8"Vibe,1,46,132,10,0,48,64,23,98,3,86" },
    { u8"0173G03", u8"Harmonizer,1,64,64,64,0,6000,0,0,0,64,85,0" },
    { u8"0173G04", u8"Valve,1,9,64,64,80,64,1,5688,220,1,1,58,1,30" },
    { u8"0173G05", u8"EQ,1,69,71,71,64,65,67,72,65,58,40,64,55" },
    { u8"0173G06", u8"Echoverse,1,64,64,90,64,64,64,64,0,2,96" },
    { u8"0174G",   u8"Dirty Junk" },
    { u8"0174G00", u8"EQ,1,64,64,64,64,65,72,64,64,64,64,64,30" },
    { u8"0174G01", u8"Compressor,1,-15,3,0,106,120,1,71,0,1" },
    { u8"0174G02", u8"DistBand,1,14,64,0,65,53,26,28,26,38,45,35,0,341,1004,0" },
    { u8"0174G03", u8"Distorsion,1,0,64,35,7,60,28,0,3327,143,0,1,0,0" },
    { u8"0175G",   u8"Jazz Hall" },
    { u8"0175G00", u8"EQ,1,64,64,66,64,61,68,67,64,64,64,64,30" },
    { u8"0175G01", u8"Compressor,1,-11,6,0,10,10,0,71,0,0" },
    { u8"0175G02", u8"Distorsion,1,55,64,64,12,55,28,1,3327,706,0,1,0,0" },
    { u8"0175G03", u8"Reverbtron,1,88,0,1,400,0,0,88,14,5,0,64,64,1,0,5216,0" },
    { u8"0175G04", u8"Cabinet,1,8,58" },
    { u8"0176G",   u8"Jazz Bright Chorus" },
    { u8"0176G00", u8"EQ,1,64,67,66,64,61,68,67,64,64,64,64,12" },
    { u8"0176G01", u8"DualFlange,1,-13,0,0,37,6,0,17,4743,0,0,88,20,0,9" },
    { u8"0176G02", u8"Compressor,1,-11,6,0,10,10,0,71,0,0" },
    { u8"0176G03", u8"Distorsion,1,55,64,64,12,55,28,1,3327,706,0,1,0,0" },
    { u8"0176G04", u8"Reverbtron,1,88,0,1,400,0,0,88,14,5,0,64,64,1,0,5216,0" },
    { u8"0176G05", u8"Cabinet,1,4,58" },
    { u8"0177G",   u8"Trippy Harmo" },
    { u8"0177G00", u8"Shifter,1,0,64,64,900,200,-11,3,1,0,0" },
    { u8"0177G01", u8"Synthfilter,1,20,69,88,0,6,0,76,-50,6,2,1,0,19,114,221,127" },
    { u8"0178G",   u8"Mutation" },
    { u8"0178G00", u8"Sustainer,1,92,58" },
    { u8"0178G01", u8"OpticalTrem,1,127,120,10,5,0,64" },
    { u8"0178G02", u8"Synthfilter,1,0,20,60,0,0,95,122,-40,4,4,0,32,-32,500,100,0" },
    { u8"0179G",   u8"Jazz Vibrato" },
    { u8"0179G00", u8"EQ,1,64,67,66,64,61,68,67,64,64,64,64,12" },
    { u8"0179G01", u8"OpticalTrem,1,0,251,10,1,64,64" },
    { u8"0179G02", u8"Compressor,1,-11,6,0,10,10,0,71,0,0" },
    { u8"0179G03", u8"Distorsion,1,55,64,64,12,55,28,1,3327,706,0,1,0,0" },
    { u8"0179G04", u8"Reverbtron,1,88,0,1,400,0,0,88,14,5,0,64,64,1,0,5216,0" },
    { u8"0179G05", u8"Cabinet,1,4,58" },
    { u8"0180G",   u8"Reverse delay ambience" },
    { u8"0180G00", u8"EQ,1,64,67,66,64,61,68,67,64,64,64,64,12" },
    { u8"0180G01", u8"Compressor,1,-29,10,-15,45,55,1,100,0,0" },
    { u8"0180G02", u8"DualFlange,1,-13,0,0,37,6,85,17,4743,0,0,88,20,0,9" },
    { u8"0180G03", u8"Echoverse,1,69,64,12,64,98,76,23,127,0,96" },
    { u8"0180G04", u8"Reverbtron,1,88,0,1,400,0,0,88,14,5,0,64,64,1,0,5216,0" },
    { u8"0181G",   u8"PongChor" },
    { u8"0181G00", u8"EQ,1,64,67,66,64,61,68,67,64,64,64,64,12" },
    { u8"0181G01", u8"Compressor,1,-29,17,-15,45,55,1,100,0,0" },
    { u8"0181G02", u8"DualFlange,1,-13,0,0,37,6,85,17,4743,0,0,88,20,8,9" },
    { u8"0181G03", u8"Echoverse,1,69,64,12,42,98,76,23,0,3,96" },
    { u8"0181G04", u8"Echotron,1,64,61,44,7,0,81,0,64,7,32,0,64,1,1,8,1" },
    { u8"0181G05", u8"Reverbtron,1,88,0,1,400,0,0,88,14,5,0,64,64,1,0,5216,0" },
    { u8"0182G",   u8"Pitch Vibrato" },
    { u8"0182G00", u8"EQ,1,64,67,66,64,61,68,67,64,64,64,64,12" },
    { u8"0182G01", u8"Vibe,1,51,260,10,0,83,70,0,88,32,63" },
    { u8"0182G02", u8"Compressor,1,-11,6,0,10,10,0,71,0,0" },
    { u8"0182G03", u8"Distorsion,1,55,64,64,12,55,28,1,3327,706,0,1,0,0" },
    { u8"0182G04", u8"Reverbtron,1,88,0,1,400,0,0,88,14,5,0,64,64,1,0,5216,0" },
    { u8"0183G",   u8"Violin" },
    { u8"0183G00", u8"Parametric EQ,1,180,79,35,1716,75,58,8182,73,48,64" },
    { u8"0183G01", u8"EQ,1,23,14,48,58,69,60,53,51,42,53,76,60" },
    { u8"0183G02", u8"Flanger,1,64,64,96,18,0,0,9,0,16,64,0,0" },
    { u8"0183G03", u8"Chorus,1,64,64,34,0,0,90,40,85,64,119,0,0" },
    { u8"0183G04", u8"Reverbtron,1,64,0,1,400,0,0,48,20,4,0,0,64,0,0,20000,0" },
    { u8"0184G",   u8"Violin Distortion" },
    { u8"0184G00", u8"Distorsion,1,0,64,127,42,24,0,1,3134,104,0,0,0,25" },
    { u8"0184G01", u8"Parametric EQ,1,180,72,35,1716,55,53,8182,37,60,64" },
    { u8"0184G02", u8"Flanger,1,64,64,99,18,0,0,9,0,16,64,0,0" },
    { u8"0184G03", u8"Chorus,1,85,63,1,0,0,73,120,86,64,127,0,0" },
    { u8"0184G04", u8"Reverbtron,1,64,0,1,400,0,0,48,20,4,0,0,64,0,0,20000,0" },
    { u8"0185G",   u8"Delayed Chorus" },
    { u8"0185G00", u8"CompBand,1,48,16,42,42,16,-16,-10,-11,-21,120,1000,5762,64" },
    { u8"0185G01", u8"Musical Delay,1,32,0,1,7,64,68,0,127,2,68,120,103,107" },
    { u8"0185G02", u8"StereoHarm,1,48,64,12,280,64,12,-280,0,0,0,0,64" },
    { u8"0185G03", u8"Compressor,1,-60,24,0,10,500,1,48,1,1" },
    { u8"0186G",   u8"Phasetron" },
    { u8"0186G00", u8"Phaser,1,0,64,6,0,4,0,100,16,1,64,1,21" },
    { u8"0186G01", u8"StereoHarm,1,64,64,12,280,64,12,-280,0,0,0,0,64" },
    { u8"0186G02", u8"Echotron,1,64,47,28,8,0,12,0,64,3,0,0,64,1,1,0,1" },
    { u8"0186G03", u8"Pan,1,64,64,48,0,0,0,64,1,1" },
    { u8"0187G",   u8"JZ Power" },
    { u8"0187G00", u8"Sequence,1,30,127,30,50,80,40,110,80,0,240,95,1,1,2,2" },
    { u8"0187G01", u8"MuTroMojo,1,16,10,60,0,0,64,0,40,10,32,-16,40,-3,1,2000,400,1,1" },
    { u8"0187G02", u8"Musical Delay,1,64,0,2,7,0,59,0,127,4,59,120,75,75" },
    { u8"0187G03", u8"Compressor,1,-4,2,0,20,411,1,0,1,1" },
    { u8"0188G",   u8"Trigger Chorus" },
    { u8"0188G00", u8"Shuffle,1,64,13,19,-21,22,350,1200,2000,6000,-21,1" },
    { u8"0188G01", u8"Shifter,1,64,64,64,250,100,-10,0,0,0,25" },
    { u8"0188G02", u8"Echo,1,67,75,955,64,30,59,10,0,0" },
    { u8"0189G",   u8"EchoPhase" },
    { u8"0189G00", u8"Echo,1,81,60,472,100,127,67,36,0,0" },
    { u8"0189G01", u8"Phaser,1,64,64,10,0,0,88,40,64,3,0,0,20" },
    { u8"0189G02", u8"StereoHarm,1,64,64,12,280,64,12,-280,0,0,0,0,64" },
  }} };
#endif // SHR3D_SFX_CORE_RAKARRACK_PRESETS
  }
#endif // SHR3D_SFX

  {
    for (const std::pair<std::u8string, std::map<std::u8string, std::u8string>>& pair : serializedSaves)
    {
      if (pair.first == u8"Tones")
      {
#ifdef SHR3D_SFX
        loadSfxTone(pair.second);
#endif // SHR3D_SFX
      }
    }
  }

#ifdef SHR3D_SFX
  { // Load Tone Assignment 0
    for (i32 i = 0; i < ARRAY_SIZE(Global::effectChain); ++i)
    {
      Global::effectChain[i] = Global::sfxTone[0][i];
      if (Global::effectChain[i] != 0)
      {
        i32 instance = 0;
        for (i32 j = 0; j < i; ++j)
          if (Global::effectChain[i] == Global::effectChain[j])
            ++instance;

        Sfx::loadParameters(abs(Global::effectChain[i]), instance, Global::sfxParameters[0][i]);
      }
    }
  }
#endif // SHR3D_SFX
}

void Tones::fini()
{
#ifdef SHR3D_SFX
  if (Sfx::names.size() == 0)
    return;

  std::map<std::u8string, std::map<std::u8string, std::u8string>> serializedSaves;

  std::map<std::u8string, std::u8string> serializedSfxTones;
  for (auto& [toneIndex, effectChain] : Global::sfxTone)
  {
    if (toneIndex < 0)
      continue;

    char8_t keyStr[16];

    if (!Global::sfxToneNames[toneIndex].empty())
    {
      sprintf(reinterpret_cast<char*>(keyStr), "%d,%d", toneIndex / Const::sfxToneTonesPerBank, toneIndex % Const::sfxToneTonesPerBank);
      serializedSfxTones.insert({ keyStr, Global::sfxToneNames[toneIndex] });
    }

    for (i32 j = 0; j < ARRAY_SIZE(Global::effectChain); ++j)
    {
      if (effectChain[j] == 0)
        continue;

      const bool isSfxMuted = effectChain[j] < 0;
      const i32 index = abs(effectChain[j]);

      sprintf(reinterpret_cast<char*>(keyStr), "%d,%d,%d", toneIndex / Const::sfxToneTonesPerBank, toneIndex % Const::sfxToneTonesPerBank, j);

#ifdef SHR3D_SFX_CORE_RAKARRACK
      if (index < to_underlying_(SfxCore::COUNT))
      {
        const std::u8string& parameters = Global::sfxParameters[toneIndex][j];
        serializedSfxTones.insert({ keyStr, std::u8string(Sfx::index2typeName(index)) + u8',' + Sfx::names[index] + u8',' + (isSfxMuted ? u8'0' : u8'1') + u8',' + parameters });
      }
      else
#endif // SHR3D_SFX_CORE_RAKARRACK
      {
        const std::u8string& parametersBase64 = Base64::encode(reinterpret_cast<const u8*>(Global::sfxParameters[toneIndex][j].data()), Global::sfxParameters[toneIndex][j].size());
        serializedSfxTones.insert({ keyStr, std::u8string(Sfx::index2typeName(index)) + u8',' + Sfx::names[index] + u8',' + (isSfxMuted ? u8'0' : u8'1') + u8',' + parametersBase64 });
      }
    }
  }
  serializedSaves.insert({ u8"Tones", serializedSfxTones });

  Ini::saveIniFile(Settings::pathTonesIni.c_str(), serializedSaves);
#endif // SHR3D_SFX
}
