// SPDX-License-Identifier: Unlicense

#include "window.h"

#include "global.h"
#include "input.h"
#include "opengl.h"
#include "type.h"

#ifdef SHR3D_WINDOW_WIN32
#include <Windows.h>
#endif // SHR3D_WINDOW_WIN32
#ifdef SHR3D_WINDOW_SDL
#include <SDL.h>
#endif // SHR3D_WINDOW_SDL

#ifdef SHR3D_WINDOW_SDL
void Window_::toggleFullscreen(SDL_Window* hwnd, FullscreenMode& fullscreenMode)
{
  switch (fullscreenMode)
  {
  case FullscreenMode::windowed:
    fullscreenMode = FullscreenMode::borderless;
    SDL_SetWindowFullscreen(hwnd, SDL_WINDOW_FULLSCREEN_DESKTOP);
    break;
  case FullscreenMode::borderless:
    fullscreenMode = FullscreenMode::windowed;
    SDL_SetWindowFullscreen(hwnd, 0);
    break;
  default:
    unreachable();
  }
}

void Window_::handleInput(SDL_Event& event)
{
  switch (event.type)
  {
  case SDL_QUIT:
    Global::appQuit = true;
    break;
  case SDL_KEYDOWN:
#ifdef SHR3D_COOP
    if (event.window.windowID == Global::coopWindowId)
    {
      if (event.key.keysym.sym == SDLK_F4)
      {
        const i32 manifestIndexCount = Global::songTracks.size();

        if (manifestIndexCount >= 1)
        {
          Global::coopSelectedManifestIndex = (Global::coopSelectedManifestIndex + 1) % manifestIndexCount;

          if (Global::songInfos[Global::selectedSongIndex].manifestInfos[Global::coopSelectedManifestIndex].arrangementName == u8"Vocals")
            Global::coopSelectedManifestIndex = (Global::coopSelectedManifestIndex + 1) % manifestIndexCount;
        }
      }
    }
    else
#endif // SHR3D_COOP
    {
      Input::keyStateChange(event.key.keysym.sym, true);
    }
    break;
  case SDL_KEYUP:
#ifdef SHR3D_COOP
    if (event.window.windowID == Global::coopWindowId)
    {
    }
    else
#endif // SHR3D_COOP
    {
      Input::keyStateChange(event.key.keysym.sym, false);
    }
    break;
  case SDL_MOUSEBUTTONDOWN:
    Input::keyStateChange(event.button.button, true);
    break;
  case SDL_MOUSEBUTTONUP:
    Input::keyStateChange(event.button.button, false);
    break;
  case SDL_MOUSEMOTION:
    Global::inputUseController = false;
    Global::inputCursorPosX = event.motion.x;
    Global::inputCursorPosY = event.motion.y;
    break;
  case SDL_CONTROLLERBUTTONDOWN:
    switch (event.cbutton.button)
    {
    case SDL_CONTROLLER_BUTTON_A:
      break;
    case SDL_CONTROLLER_BUTTON_B:
      break;
    case SDL_CONTROLLER_BUTTON_X:
      break;
    case SDL_CONTROLLER_BUTTON_Y:
      break;
    case SDL_CONTROLLER_BUTTON_RIGHTSTICK:
      break;
    case SDL_CONTROLLER_BUTTON_DPAD_LEFT:
      break;
    case SDL_CONTROLLER_BUTTON_DPAD_RIGHT:
      break;
    }
    break;
  case SDL_CONTROLLERBUTTONUP:
    switch (event.cbutton.button)
    {
    case SDL_CONTROLLER_BUTTON_A:
      break;
    case SDL_CONTROLLER_BUTTON_B:
      break;
    case SDL_CONTROLLER_BUTTON_X:
      break;
    case SDL_CONTROLLER_BUTTON_Y:
      break;
    case SDL_CONTROLLER_BUTTON_RIGHTSTICK:
      break;
    case SDL_CONTROLLER_BUTTON_DPAD_LEFT:
      break;
    case SDL_CONTROLLER_BUTTON_DPAD_RIGHT:
      break;
    }
    break;
  case SDL_JOYAXISMOTION:
    switch (event.caxis.axis)
    {
    case SDL_CONTROLLER_AXIS_LEFTX:
      break;
    case SDL_CONTROLLER_AXIS_LEFTY:
      break;
    case SDL_CONTROLLER_AXIS_RIGHTX:
      break;
    case SDL_CONTROLLER_AXIS_RIGHTY:
      break;
    case SDL_CONTROLLER_AXIS_TRIGGERLEFT:
      break;
    case SDL_CONTROLLER_AXIS_TRIGGERRIGHT:
      break;
    }
    break;
  case SDL_CONTROLLERDEVICEADDED:
    if (Global::gameController == nullptr && SDL_IsGameController(0))
      Global::gameController = SDL_GameControllerOpen(0);
    break;
  case SDL_CONTROLLERDEVICEREMOVED:
    SDL_GameControllerClose(Global::gameController);
    Global::gameController = nullptr;
    break;
  case SDL_WINDOWEVENT:
  {
    switch (event.window.event)
    {
#ifdef SHR3D_COOP
    case SDL_WINDOWEVENT_CLOSE:
      if (event.window.windowID == Global::coopWindowId)
        Global::inputCoop.toggled = !Global::inputCoop.toggled;
      else
        Global::appQuit = true;
      break;
#endif // SHR3D_COOP
    case SDL_WINDOWEVENT_SIZE_CHANGED:
#ifdef SHR3D_COOP
      if (event.window.windowID == Global::coopWindowId)
      {
        Global::coopResolutionWidth = event.window.data1;
        Global::coopResolutionHeight = event.window.data2;
      }
      else
#endif // SHR3D_COOP
      {
        Global::resolutionWidth = event.window.data1;
        Global::resolutionHeight = event.window.data2;

        if (Settings::graphicsFullscreen == FullscreenMode::windowed)
        {
          Settings::graphicsWindowWidth = Global::resolutionWidth;
          Settings::graphicsWindowHeight = Global::resolutionHeight;
        }
        GL(glViewport(0, 0, Global::resolutionWidth, Global::resolutionHeight));
      }
      break;
    }
  }
  break;
  }
}
#endif // SHR3D_WINDOW_SDL

#ifdef SHR3D_WINDOW_WIN32
void Window_::toggleFullscreen(const HWND hwnd, const i32 resolutionWidth, const i32 resolutionHeight, FullscreenMode& fullscreenMode, WINDOWPLACEMENT& lastWindowPlacement)
{
  if (fullscreenMode == FullscreenMode::windowed)
  {
    fullscreenMode = FullscreenMode::borderless;

    MONITORINFO mi = { sizeof(mi) };
    GetWindowPlacement(hwnd, &lastWindowPlacement);
    if (GetMonitorInfo(MonitorFromWindow(hwnd, MONITOR_DEFAULTTOPRIMARY), &mi)) {

      SetWindowLong(hwnd, GWL_STYLE, 335544320);
      SetWindowPos(hwnd, HWND_TOP,
        mi.rcMonitor.left, mi.rcMonitor.top,
        mi.rcMonitor.right - mi.rcMonitor.left,
        mi.rcMonitor.bottom - mi.rcMonitor.top,
        SWP_NOOWNERZORDER | SWP_FRAMECHANGED);
    }
  }
  else
  {
    if (lastWindowPlacement.length == 0) // lastWindowPlacement was not initialized, then initialize it to the center of screen and the known window width/height
    {
      RECT rect = { 0, 0, resolutionWidth, resolutionHeight };
      AdjustWindowRectEx(&rect, WS_OVERLAPPEDWINDOW, FALSE, 0);
      const i32 windowWidth = rect.right - rect.left;
      const i32 windowHeight = rect.bottom - rect.top;

      lastWindowPlacement = WINDOWPLACEMENT{
        .length = sizeof(lastWindowPlacement),
        .showCmd = 1,
        .ptMinPosition {
          .x = -1,
          .y = -1
        },
        .ptMaxPosition {
          .x = -1,
          .y = -1
        },
        .rcNormalPosition {
          .left = (GetSystemMetrics(SM_CXSCREEN) - windowWidth) / 2,
          .top = (GetSystemMetrics(SM_CYSCREEN) - windowHeight) / 2,
          .right = (GetSystemMetrics(SM_CXSCREEN) + windowWidth) / 2,
          .bottom = (GetSystemMetrics(SM_CYSCREEN) + windowHeight) / 2,
        }
      };
    }

    fullscreenMode = FullscreenMode::windowed;
    SetWindowLong(hwnd, GWL_STYLE, 349110272);
    SetWindowPlacement(hwnd, &lastWindowPlacement);
    SetWindowPos(hwnd, NULL, 0, 0, 0, 0,
      SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER |
      SWP_NOOWNERZORDER | SWP_FRAMECHANGED);
  }
}
#endif // SHR3D_WINDOW_WIN32
